import logging

from nabu.resources.nxflatfield import update_dataset_info_flats_darks
from nabu.resources.dataset_analyzer import HDF5DatasetAnalyzer
from nabu.io.reader import load_images_from_dataurl_dict
from nabu.app.prepare_weights_double import create_heli_maps

from tomwer.core.process.task import TaskWithProgress
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.utils.scanutils import format_output_location


_logger = logging.getLogger(__name__)


class NabuHelicalPrepareWeightsDouble(
    TaskWithProgress,
    input_names=(
        "data",
        "transition_width_vertical",
        "transition_width_horizontal",
        "processes_file",
        "rotation_axis_position",
    ),
    optional_input_names=("progress",),
    output_names=("data",),
):
    def run(self):
        # TODO: handle future /cluster config ???
        scan = self.inputs.data
        if not isinstance(scan, NXtomoScan):
            raise TypeError(f"data is expected to be an instance of {NXtomoScan}")
        dataset_info = HDF5DatasetAnalyzer(
            scan.master_file,
            extra_options={"h5_entry": scan.entry},
        )
        update_dataset_info_flats_darks(dataset_info, flatfield_mode=1)

        mappe = 0
        my_flats = load_images_from_dataurl_dict(dataset_info.flats)

        for _, flat in my_flats.items():
            mappe += flat
        mappe = mappe / len(list(dataset_info.flats.keys()))

        scan.helical.processes_files = format_output_location(
            location=self.inputs.processes_file, scan=scan
        )
        create_heli_maps(
            profile=mappe,
            process_file_name=scan.helical.processes_files,
            entry_name=scan.entry,
            transition_width_vertical=self.inputs.transition_width_vertical,
            transition_width_horizontal=self.inputs.transition_width_horizontal,
            rotation_axis_position=self.inputs.rotation_axis_position,
        )

        self.outputs.data = scan



# coding: utf-8
from __future__ import annotations

import pytest
from silx.gui import qt
from silx.gui.utils.testutils import SignalListener, TestCaseQt

from tomwer.gui.control.datalistener import ConfigurationWidget
from tomwer.tests.utils import skip_gui_test


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestDataListenerConfiguration(TestCaseQt):
    """
    Simple test the interface for configuration is working
    """

    def setUp(self):
        TestCaseQt.setUp(self)
        self._configWidget = ConfigurationWidget(parent=None)
        self.sig_listener = SignalListener()
        self._configWidget.sigConfigurationChanged.connect(self.sig_listener)
        self._configWidget.setHost("localhost")

    def tearDown(self):
        self._configWidget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._configWidget.close()
        self.qapp.processEvents()
        self._configWidget = None

    def testConfiguration(self):
        self.assertEqual(
            self._configWidget.getConfiguration(), {"host": "localhost", "port": 4000}
        )
        self._configWidget.setPort(0)
        self._configWidget.setHost("toto")
        self.assertEqual(
            self._configWidget.getConfiguration(), {"host": "toto", "port": 0}
        )

    def testInteraction(self):
        """make sure the sigConfigurationChanged is emitted only when the
        user validate the modifications. This is done because we need to
        stop the thread and launch a new one, we don't wan't to do this
        automatically"""
        self._configWidget.setPort(4000)
        self._configWidget.setHost("tata")
        self.assertEqual(self.sig_listener.callCount(), 0)
        self._configWidget.validate()
        self.qapp.processEvents()
        self.assertEqual(self.sig_listener.callCount(), 1)



# coding: utf-8
from __future__ import annotations

import gc
import logging
import os
import shutil
import tempfile
import time
from time import sleep

from silx.gui import qt
from silx.gui.utils.testutils import SignalListener, TestCaseQt

from orangecontrib.tomwer.tests.TestAcquisition import Simulation
from orangecontrib.tomwer.widgets.control.DataWatcherOW import DataWatcherOW
from tomwer.core.process.control.datawatcher import status as datawatcherstatus
from tomwer.core.utils.scanutils import MockBlissAcquisition, MockNXtomo

logging.disable(logging.INFO)


class DataWatcherWaiter(TestCaseQt):
    """Define a simple objecy able to wait for some state of the DataWatcher
    arrive"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.reset()
        self.lastStatus = []

    def tearDown(self):
        pass

    def reset(self):
        self.lastStatus = []

    def stateHasChanged(self, val):
        """Register all status"""
        if val not in self.lastStatus:
            self.lastStatus.append(val)

    def waitForState(self, state, maxWaiting):
        """simple function wich wait until the DataWatcherWidget reach the given
        state.
        If the widget doesn't reach this state after maxWaiting second. Then fail.

        :param state: the state we are waiting for
        :param maxWaiting: the maximal number of second to wait until failling.
        """
        while state not in self.lastStatus:
            time.sleep(1.0)
            self.qapp.processEvents()
            maxWaiting -= 1
            if maxWaiting <= 0:
                return False
        return state in self.lastStatus


class TestDataWatcherAcquisition(DataWatcherWaiter):
    """Functions testing the classical behavior of data watcher
    - signal acquisition is over only when all files are copied
    """

    def tearDown(self):
        while self.qapp.hasPendingEvents():
            self.qapp.processEvents()
        self.dataWatcherWidget.setObservation(False)
        self.dataWatcherWidget.close()
        del self.dataWatcherWidget
        self.dataWatcherWidget = None
        self.s.wait()
        del self.s
        if os.path.isdir(self.inputdir):
            shutil.rmtree(self.inputdir)
        gc.collect()
        super().tearDown()

    def setUp(self):
        self.manipulationId = "test10"

        super().setUp()
        self.inputdir = tempfile.mkdtemp()
        DataWatcherWaiter.reset(self)
        self.dataWatcherWidget = DataWatcherOW(displayAdvancement=False)
        self.dataWatcherWidget.srcPattern = ""
        self.dataWatcherWidget.sigTMStatusChanged.connect(self.stateHasChanged)
        self.dataWatcherWidget.setAttribute(qt.Qt.WA_DeleteOnClose)

        self.s = Simulation(
            self.inputdir,
            self.manipulationId,
            finalState=Simulation.advancement["acquisitionRunning"],
        )

    def testStartAcquisition(self):
        """Make sure the data watch detect the acquisition of started"""
        observeDir = os.path.join(self.inputdir, self.manipulationId)
        for folder in (self.inputdir, observeDir):
            if not os.path.exists(folder):
                os.makedirs(folder)

        self.assertTrue(os.path.isdir(observeDir))

        self.s.createFinalXML(True)
        self.dataWatcherWidget.setFolderObserved(observeDir)

        self.assertTrue(self.dataWatcherWidget.currentStatus == "not processing")
        self.s.start()
        self.s.advanceTo("acquisitionDone")
        self.s.start()
        self.s.wait()
        self.dataWatcherWidget.startObservation()
        self.dataWatcherWidget._widget.observationThread.wait()
        self.dataWatcherWidget._widget.observationThread.observations.dict[
            observeDir
        ].wait()
        self.qapp.processEvents()
        finishedAcqui = (
            self.dataWatcherWidget._widget.observationThread.observations.ignoredFolders
        )
        self.qapp.processEvents()
        self.assertTrue(observeDir in finishedAcqui)


class TestDataWatcherInteraction(TestCaseQt):
    """Simple unit test to test the start/stop observation button action"""

    def setUp(self):
        super().setUp()
        self.inputdir = tempfile.mkdtemp()
        self.dataWatcherWidget = DataWatcherOW(displayAdvancement=False)
        self.dataWatcherWidget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.dataWatcherWidget.srcPattern = ""

    def tearDown(self):
        while self.qapp.hasPendingEvents():
            self.qapp.processEvents()
        self.dataWatcherWidget.close()
        del self.dataWatcherWidget
        if os.path.isdir(self.inputdir):
            shutil.rmtree(self.inputdir)
        gc.collect()

    def testStartAndStopAcquisition(self):
        """test multiple start and stop action on the start observation to
        make sure no crash are appearing
        """
        try:
            self.dataWatcherWidget.setFolderObserved(self.inputdir)
            self.dataWatcherWidget.show()
            self.dataWatcherWidget.setObservation(True)
            for _ in range(5):
                self.dataWatcherWidget._widget._qpbstartstop.pressed.emit()
            self.assertTrue(True)
        except Exception:
            self.assertTrue(False)


class WaitForXMLOption(DataWatcherWaiter):
    """test the behavior of datawatcher when the option 'wait for xml copy'
    Basically in this case DataWatcherDirObserver will wait until an .xml
    arrived
    """

    @classmethod
    def setUpClass(cls):
        cls.dataWatcherWidget = DataWatcherOW(displayAdvancement=False)
        cls.dataWatcherWidget.setAttribute(qt.Qt.WA_DeleteOnClose)
        cls.dataWatcherWidget.obsMethod = datawatcherstatus.DET_END_XML
        cls.dataWatcherWidget.srcPattern = ""
        cls.manipulationId = "test10"
        super().setUpClass()

    def setUp(self):
        self.inputdir = tempfile.mkdtemp()
        self.reset()
        self.dataWatcherWidget.setObservation(False)
        self.dataWatcherWidget.resetStatus()
        super().setUp()

    def tearDown(self):
        while self.qapp.hasPendingEvents():
            self.qapp.processEvents()
        if os.path.isdir(self.inputdir):
            shutil.rmtree(self.inputdir)
        gc.collect()
        super().tearDown()

    @classmethod
    def tearDownClass(cls):
        cls.dataWatcherWidget.close()
        del cls.dataWatcherWidget
        if hasattr(cls, "s"):
            cls.s.quit()
            del cls.s
        super().tearDownClass()

    def testAcquistionNotEnding(self):
        """Check behavior if an acquisition never end"""
        observeDir = os.path.join(self.inputdir, self.manipulationId)
        for folder in (self.inputdir, observeDir):
            if not os.path.exists(folder):
                os.makedirs(folder)

        self.assertTrue(os.path.isdir(observeDir))

        self.s = Simulation(
            self.inputdir,
            self.manipulationId,
            finalState=Simulation.advancement["acquisitionRunning"],
        )
        self.dataWatcherWidget.setFolderObserved(observeDir)
        self.dataWatcherWidget.show()
        self.dataWatcherWidget.sigTMStatusChanged.connect(self.stateHasChanged)

        self.assertTrue(self.dataWatcherWidget.currentStatus == "not processing")
        self.s.start()
        self.s.wait()
        self.dataWatcherWidget.setObservation(True)
        self.dataWatcherWidget._widget.observationThread.wait()
        self.dataWatcherWidget._widget.observationThread.observations.dict[
            observeDir
        ].wait()
        finishedAcqui = (
            self.dataWatcherWidget._widget.observationThread.observations.ignoredFolders
        )
        self.qapp.processEvents()
        self.assertFalse(observeDir in finishedAcqui)

    def testAcquistionEnded(self):
        """Check behavior if an acquisition is ending"""
        manipulationId = "test10"
        observeDir = os.path.join(self.inputdir, self.manipulationId)
        for folder in (self.inputdir, observeDir):
            if not os.path.exists(folder):
                os.makedirs(folder)

        self.assertTrue(os.path.isdir(observeDir))

        self.s = Simulation(
            self.inputdir,
            manipulationId,
            finalState=Simulation.advancement["acquisitionDone"],
        )
        self.s.createFinalXML(True)
        self.dataWatcherWidget.setFolderObserved(observeDir)
        self.dataWatcherWidget.show()
        self.dataWatcherWidget.sigTMStatusChanged.connect(self.stateHasChanged)

        self.assertTrue(self.dataWatcherWidget.currentStatus == "not processing")
        self.s.start()
        self.s.wait()
        self.dataWatcherWidget.setObservation(True)
        self.dataWatcherWidget._widget.observationThread.wait()
        self.dataWatcherWidget._widget.observationThread.observations.dict[
            observeDir
        ].wait()
        finishedAcqui = (
            self.dataWatcherWidget._widget.observationThread.observations.ignoredFolders
        )
        self.qapp.processEvents()
        self.assertTrue(observeDir in finishedAcqui)


class TestRSync(DataWatcherWaiter):
    """test that the synchronization using RSyncManager is working"""

    def setUp(self):
        super().setUp()
        self.inputdir = tempfile.mkdtemp()
        self.outputdir = tempfile.mkdtemp()
        DataWatcherWaiter.reset(self)
        self.dataWatcherWidget = DataWatcherOW(displayAdvancement=False)
        self.dataWatcherWidget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.dataWatcherWidget._widget.setSrcAndDestPattern(
            self.inputdir, self.outputdir
        )

    def tearDown(self):
        while self.qapp.hasPendingEvents():
            self.qapp.processEvents()
        self.dataWatcherWidget.close()
        del self.dataWatcherWidget
        if hasattr(self, "s"):
            self.s.quit()
            del self.s
        super(TestRSync, self).tearDown()
        for d in (self.inputdir, self.outputdir):
            if os.path.isdir(d):
                shutil.rmtree(d, ignore_errors=True)
        gc.collect()

    def testStartAcquisition(self):
        """Test that rsync is launched when an acquisition is discovered"""
        manipulationId = "test10"
        observeDir = os.path.join(self.inputdir, manipulationId)
        for folder in (self.inputdir, observeDir):
            if not os.path.exists(folder):
                os.makedirs(folder)

        self.assertTrue(os.path.isdir(observeDir))

        self.s = Simulation(
            self.inputdir,
            manipulationId,
            finalState=Simulation.advancement["acquisitionRunning"],
        )

        self.s.setSrcDestPatterns(self.inputdir, self.outputdir)
        self.s.createFinalXML(True)
        self.dataWatcherWidget.setFolderObserved(self.inputdir)
        self.dataWatcherWidget.show()
        self.dataWatcherWidget.sigTMStatusChanged.connect(self.stateHasChanged)
        self.assertTrue(self.dataWatcherWidget.currentStatus == "not processing")
        self.dataWatcherWidget.setObservation(True)
        self.s.start()
        # check state scanning
        time.sleep(0.5)

        self.dataWatcherWidget.stopObservation()

        # in this case the .info should be in the output dir also
        test10_output = os.path.join(self.outputdir, "test10")
        test10_input = os.path.join(self.inputdir, "test10")
        self.assertTrue(os.path.isfile(os.path.join(test10_output, "test10.info")))

        # make sure file transfert have been started (using rsync)
        # all file in outputdir should be in input dir
        time.sleep(2)
        # check that some .edf file have already been copied
        self.assertTrue(len(test10_output) > 5)

        # xml shouldn't be there because we are righting it at the end
        self.assertFalse(os.path.isfile(os.path.join(test10_output, "test10.xml")))
        self.assertFalse(os.path.isfile(os.path.join(test10_input, "test10.xml")))


class TestDataWatcherBlissScan(TestCaseQt):
    def setUp(self):
        self._widget = DataWatcherOW(self)
        self.tempdir = tempfile.mkdtemp()
        MockBlissAcquisition(
            n_sample=1,
            n_sequence=1,
            n_scan_per_sequence=3,
            n_darks=2,
            n_flats=2,
            output_dir=os.path.join(self.tempdir, "folder_1"),
        )

        MockBlissAcquisition(
            n_sample=2,
            n_sequence=1,
            n_scan_per_sequence=3,
            n_darks=2,
            n_flats=2,
            output_dir=os.path.join(self.tempdir, "test", "with", "some", "depth"),
        )
        self._listener = SignalListener()
        self._widget._widget.sigScanReady.connect(self._listener)

    def tearDown(self):
        self._widget._widget.sigScanReady.disconnect(self._listener)
        self._widget.stopObservation()
        self._listener = None
        shutil.rmtree(self.tempdir)
        self._widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._widget.close()
        self._widget = None
        gc.collect()


class TestDataWatcherNXtomo(TestCaseQt):
    def setUp(self):
        self._widget = DataWatcherOW(self)
        self.tempdir = tempfile.mkdtemp()
        MockNXtomo(
            scan_path=os.path.join(self.tempdir, "folder_1"),
            n_proj=10,
        )

        MockNXtomo(
            scan_path=os.path.join(self.tempdir, "test", "with", "some", "depth"),
            n_proj=10,
        )
        self._listener = SignalListener()
        self._widget._widget.sigScanReady.connect(self._listener)

    def tearDown(self):
        self._widget._widget.sigScanReady.disconnect(self._listener)
        self._widget.stopObservation()
        self._listener = None
        shutil.rmtree(self.tempdir)
        self._widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._widget.close()
        self._widget = None
        gc.collect()

    def test(self):
        self._widget.setFolderObserved(self.tempdir)
        self._widget._widget.getConfigWindow().setMode((datawatcherstatus.NXtomo_END,))
        self._widget.show()
        self._widget.startObservation()
        processing_time = 2
        sleep_time = 0.005
        while processing_time > 0:
            self.qapp.processEvents()
            sleep(sleep_time)
            processing_time -= sleep_time
        assert self._listener.callCount() == 2



# coding: utf-8
from __future__ import annotations

import enum

from silx.utils.enum import Enum as _Enum
from tomoscan.framereducer.method import ReduceMethod

from tomwer.core.process.reconstruction.darkref.settings import (
    DARKHST_PREFIX,
    REFHST_PREFIX,
)
from tomwer.core.process.reconstruction.paramsbase import _ReconsParam
from tomwer.core.utils.deprecation import deprecated


# TODO: those classes (when, method) should be linked / embedded in the DarkRef
# method
@enum.unique
class When(_Enum):
    never = (0,)
    before = (1,)
    after = (2,)


def _from_value_reduce_method_backward_compatibility(value) -> ReduceMethod:
    if value in ("none", 0, (0,)):
        return ReduceMethod.NONE
    elif value in ("average", 1, (1,)):
        return ReduceMethod.MEAN
    elif value in ("median", 2, (2,)):
        return ReduceMethod.MEDIAN
    elif value in ("first", 10, (10,)):
        return ReduceMethod.FIRST
    elif value in ("last", 11, (11,)):
        return ReduceMethod.FIRST
    else:
        return ReduceMethod.from_value(value)


class DKRFRP(_ReconsParam):
    """Settings for the calculation of dark and flat fields mean or median"""

    def __init__(self):
        _ReconsParam.__init__(self)
        self.__do_when = When.before
        self.__dark_calc = ReduceMethod.MEAN
        self.__overwrite_dark = False
        self.__remove_dark = False
        self.__dark_pattern = "darkend[0-9]{3,4}"
        self.__flat_calc = ReduceMethod.MEDIAN
        self.__overwrite_flat = False
        self.__remove_flats = False
        self.__flat_pattern = "ref*.*[0-9]{3,4}_[0-9]{3,4}"
        self.__dark_prefix = DARKHST_PREFIX
        self.__ref_prefix = REFHST_PREFIX

        self._managed_params = {
            "DOWHEN": self.__class__.do_when,
            "DARKCAL": self.__class__.dark_calc_method,
            "DARKOVE": self.__class__.overwrite_dark,
            "DARKRMV": self.__class__.remove_dark,
            "DKFILE": self.__class__.dark_pattern,
            "REFSCAL": self.__class__.flat_calc_method,
            "REFSOVE": self.__class__.overwrite_flat,
            "REFSRMV": self.__class__.remove_raw_flats,
            "RFFILE": self.__class__.flat_pattern,
        }

    @property
    def do_when(self):
        """When should we process calculation. Should be removed now that DKRF
        process exists. Was needed for fastomo3"""
        return self.__do_when

    @do_when.setter
    def do_when(self, when):
        assert isinstance(when, (int, When))
        when = When(when)
        if when != self.__do_when:
            self.__do_when = when
            self.changed()

    @property
    def dark_calc_method(self):
        """Dark calculation Method"""
        return self.__dark_calc

    @dark_calc_method.setter
    def dark_calc_method(self, method):
        assert isinstance(method, (int, ReduceMethod, str))
        _dark_calc = _from_value_reduce_method_backward_compatibility(method)

        if self.__dark_calc != _dark_calc:
            self.__dark_calc = _dark_calc
            self.changed()

    @property
    def overwrite_dark(self):
        """Overwrite Dark results if already exists"""
        return self.__overwrite_dark

    @overwrite_dark.setter
    def overwrite_dark(self, overwrite):
        assert isinstance(overwrite, (int, bool, float))
        _overwrite_dark = bool(overwrite)
        if self.__overwrite_dark != _overwrite_dark:
            self.__overwrite_dark = _overwrite_dark
            self.changed()

    @property
    def remove_dark(self):
        """Remove original Darks files when done"""
        return self.__remove_dark

    @remove_dark.setter
    def remove_dark(self, remove):
        assert isinstance(remove, (int, bool, float))
        _remove_dark = bool(remove)
        if _remove_dark != self.__remove_dark:
            self.__remove_dark = _remove_dark
            self.changed()

    @property
    def dark_pattern(self):
        """File pattern to detect edf Dark field"""
        return self.__dark_pattern

    @dark_pattern.setter
    def dark_pattern(self, pattern):
        _dark_pattern = pattern
        if self.__dark_pattern != _dark_pattern:
            self.__dark_pattern = _dark_pattern
            self.changed()

    @property
    def flat_calc_method(self):
        """Dark calculation method (None, Average, Median)"""
        return self.__flat_calc

    @flat_calc_method.setter
    def flat_calc_method(self, method):
        assert isinstance(method, (int, ReduceMethod, str))
        _ref_calc = _from_value_reduce_method_backward_compatibility(method)
        if self.__flat_calc != _ref_calc:
            self.__flat_calc = _ref_calc
            self.changed()

    @property
    @deprecated(replacement="overwrite_flat", since_version="0.9")
    def overwrite_ref(self):
        return self.overwrite_flat

    @overwrite_ref.setter
    @deprecated(replacement="overwrite_flat", since_version="0.9")
    def overwrite_ref(self, overwrite):
        self.overwrite_flat = overwrite

    @property
    def overwrite_flat(self):
        """Overwrite Dark results if already exists"""
        return self.__overwrite_flat

    @overwrite_flat.setter
    def overwrite_flat(self, overwrite):
        assert isinstance(overwrite, (int, bool, float))
        _overwrite_flat = bool(overwrite)
        if self.__overwrite_flat != _overwrite_flat:
            self.__overwrite_flat = _overwrite_flat
            self.changed()

    @property
    def remove_raw_flats(self):
        """Remove original flat / ref files when done (for EDF only)"""
        return self.__remove_flats

    @remove_raw_flats.setter
    def remove_raw_flats(self, remove):
        # TODO: float should be removed, but this is a legacy from fastomo3
        assert isinstance(remove, (int, bool, float))
        _remove_flats = remove
        if self.__remove_flats != _remove_flats:
            self.__remove_flats = _remove_flats
            self.changed()

    @property
    def flat_pattern(self):
        """File pattern to detect EDF flats files"""
        return self.__flat_pattern

    @flat_pattern.setter
    def flat_pattern(self, pattern):
        if pattern != self.__flat_pattern:
            self.__flat_pattern = pattern
            self.changed()

    @property
    def flat_prefix(self):
        """flat prefix for EDF"""
        return self.__ref_prefix

    @flat_prefix.setter
    def flat_prefix(self, prefix):
        if prefix != self.__ref_prefix:
            self.__ref_prefix = prefix
            self.changed()

    @property
    def dark_prefix(self):
        """dark prefix for EDF"""
        return self.__dark_prefix

    @dark_prefix.setter
    def dark_prefix(self, prefix):
        if prefix != self.__dark_prefix:
            self.__dark_prefix = prefix
            self.changed()

    def _set_remove_opt(self, rm):
        self.remove_raw_flats = rm
        self.remove_dark = rm

    def _set_skip_if_exist(self, skip):
        self.overwrite_flat = not skip
        self.overwrite_dark = not skip

    def to_dict(self):
        _dict = {
            "DOWHEN": self.do_when.name,
            "DARKCAL": self.dark_calc_method.value,
            "DARKOVE": int(self.overwrite_dark),
            "DARKRMV": int(self.remove_dark),
            "DKFILE": self.dark_pattern,
            "REFSCAL": self.flat_calc_method.value,
            "REFSOVE": int(self.overwrite_flat),
            "REFSRMV": int(self.remove_raw_flats),
            "RFFILE": self.flat_pattern,
        }
        _dict.update(self.unmanaged_params)
        return _dict

    @staticmethod
    def from_dict(_dict):
        params = DKRFRP()
        params.load_from_dict(_dict)
        return params

    def load_from_dict(self, _dict):
        # Convert managed keys to upper case
        _dict = {
            key.upper() if key.upper() in self.managed_params else key: value
            for key, value in _dict.items()
        }

        self._load_unmanaged_params(_dict=_dict)
        if "DOWHEN" in _dict:
            self.do_when = getattr(When, _dict["DOWHEN"])
        if "DARKCAL" in _dict:
            self.dark_calc_method = _dict["DARKCAL"]
        if "DARKOVE" in _dict:
            self.overwrite_dark = _dict["DARKOVE"]
        if "DARKRMV" in _dict:
            self.remove_dark = _dict["DARKRMV"]
        if "DKFILE" in _dict:
            self.dark_pattern = _dict["DKFILE"]
        if "REFSCAL" in _dict:
            self.flat_calc_method = _dict["REFSCAL"]
        if "REFSOVE" in _dict:
            self.overwrite_flat = _dict["REFSOVE"]
        if "REFSRMV" in _dict:
            self.remove_raw_flats = _dict["REFSRMV"]
        if "RFFILE" in _dict:
            self.flat_pattern = _dict["RFFILE"]



# coding: utf-8
from __future__ import annotations

import pytest
from silx.gui import qt
from silx.gui.utils.testutils import SignalListener, TestCaseQt

from tomwer.gui.control.datalistener import ConfigurationWidget
from tomwer.tests.utils import skip_gui_test


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestProcessManager(TestCaseQt):
    """
    Simple test on behavior of the ProcessManager
    """

    def setUp(self):
        TestCaseQt.setUp(self)
        self._configWidget = ConfigurationWidget(parent=None)
        self.sig_listener = SignalListener()
        self._configWidget.sigConfigurationChanged.connect(self.sig_listener)
        self._configWidget.setHost("localhost")

    def tearDown(self):
        self._configWidget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._configWidget.close()
        self._configWidget = None

    def testConfiguration(self):
        self.assertEqual(
            self._configWidget.getConfiguration(), {"host": "localhost", "port": 4000}
        )
        self._configWidget.setPort(0)
        self._configWidget.setHost("toto")
        self.assertEqual(
            self._configWidget.getConfiguration(), {"host": "toto", "port": 0}
        )



# coding: utf-8
from __future__ import annotations

import gc
import logging
import os
import pickle
import shutil
import tempfile

from orangecanvas.scheme.readwrite import literal_dumps
from processview.core.manager import DatasetState, ProcessManager
from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt
from silx.io.utils import h5py_read_dataset

from orangecontrib.tomwer.widgets.reconstruction.SinoNormOW import (
    SinoNormOW as _NormIOW,
)
from tomwer.core import settings
from tomwer.core.utils.lbsram import mock_low_memory
from tomwer.core.process.reconstruction.normalization import SinoNormalizationTask
from tomwer.core.utils.scanutils import MockNXtomo
from tomwer.io.utils.h5pyutils import EntryReader

logger = logging.getLogger(__name__)


class NormIOW(_NormIOW):
    def __init__(self, parent=None):
        self._scans_finished = []
        super().__init__(parent)

    def processing_finished(self, scan):
        self._scans_finished.append(scan)

    def wait_processing(self, wait_time):
        self._window._processing_stack._computationThread.wait(wait_time)

    @property
    def scans_finished(self):
        return self._scans_finished

    def compute(self):
        self._window._processCurrentScan()

    def setROI(self, start_x, end_x, start_y, end_y):
        self._window.setROI(start_x=start_x, end_x=end_x, start_y=start_y, end_y=end_y)

    def close(self):
        self._scans_finished = {}
        super().close()


class TestProcessing(TestCaseQt):
    DIM = 100

    def setUp(self):
        super().setUp()
        self._source_dir = tempfile.mkdtemp()

        def create_scan(folder_name):
            _dir = os.path.join(self._source_dir, folder_name)
            return MockNXtomo(
                scan_path=_dir,
                n_ini_proj=20,
                n_proj=20,
                n_alignement_proj=2,
                create_final_flat=False,
                create_ini_dark=True,
                create_ini_flat=True,
                n_refs=1,
                dim=self.DIM,
            ).scan

        # create scans
        self.scan_1 = create_scan("scan_1")
        self.scan_2 = create_scan("scan_2")
        self.scan_3 = create_scan("scan_3")
        self._process_manager = ProcessManager()

        self.widget = NormIOW()
        self.widget.show()

    def tearDown(self):
        mock_low_memory(False)
        settings.mock_lsbram(False)
        self.widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.widget.close()
        self.widget = None
        self.qapp.processEvents()
        shutil.rmtree(self._source_dir)
        gc.collect()

    def test_serializing(self):
        pickle.dumps(self.widget.getConfiguration())

    def test_literal_dumps(self):
        self.widget._updateSettings()
        literal_dumps(self.widget._ewoks_default_inputs)

    def testUnlocked(self):
        """Test result when used with some interaction"""
        self.widget.setLocked(False)

        def process_scalar_manually():
            self.widget.setCurrentMethod("division")
            self.widget.setCurrentSource("manual ROI")

            self.qapp.processEvents()
            self.widget.setROI(start_x=0, end_x=10, start_y=0, end_y=10)
            self.qapp.processEvents()
            self.widget.compute()
            self.widget.wait_processing(5000)
            self.qapp.processEvents()

        self.widget.process(self.scan_1)
        process_scalar_manually()
        self.assertEqual(
            self._process_manager.get_dataset_state(
                dataset_id=self.scan_1.get_identifier(),
                process=self.widget,
            ),
            DatasetState.WAIT_USER_VALIDATION,
        )

        self.widget.process(self.scan_2)
        process_scalar_manually()
        self.assertEqual(len(self.widget.scans_finished), 0)
        self.assertEqual(
            self._process_manager.get_dataset_state(
                dataset_id=self.scan_1.get_identifier(),
                process=self.widget,
            ),
            DatasetState.SKIPPED,
        )
        self.assertEqual(
            self._process_manager.get_dataset_state(
                dataset_id=self.scan_2.get_identifier(),
                process=self.widget,
            ),
            DatasetState.WAIT_USER_VALIDATION,
        )

        self.widget.process(self.scan_3)
        process_scalar_manually()
        self.widget.validateCurrentScan()
        self.assertEqual(
            self._process_manager.get_dataset_state(
                dataset_id=self.scan_3.get_identifier(),
                process=self.widget,
            ),
            DatasetState.SUCCEED,
        )

    def testTestLbsram(self):
        """Test scan are all validated if 'low memory on lbsram' scenario is
        activated"""
        mock_low_memory(True)
        settings.mock_lsbram(True)
        for scan in (self.scan_1, self.scan_2, self.scan_3):
            self.widget.process(scan)
            self.widget.wait_processing(5000)
            self.qapp.processEvents()

        for scan in (self.scan_1, self.scan_2, self.scan_3):
            with self.subTest(scan=str(scan)):
                self.assertEqual(
                    self._process_manager.get_dataset_state(
                        dataset_id=scan.get_identifier(),
                        process=self.widget,
                    ),
                    DatasetState.SKIPPED,
                )

    def testLocked(self):
        """Test scan are all validated if the widget is lock"""
        self.widget.setLocked(True)
        for scan in (self.scan_1, self.scan_2, self.scan_3):
            self.widget.process(scan)
            self.widget.wait_processing(5000)
            self.qapp.processEvents()

        for scan in (self.scan_1, self.scan_2, self.scan_3):
            # test status is SUCCEED
            with self.subTest(scan=str(scan)):
                self.assertEqual(
                    self._process_manager.get_dataset_state(
                        dataset_id=scan.get_identifier(),
                        process=self.widget,
                    ),
                    DatasetState.SUCCEED,
                )
            # test process file has been updated
            with EntryReader(scan.process_file_url) as entry:
                self.assertTrue("tomwer_process_0" in entry)
                self.assertEqual(
                    h5py_read_dataset(entry["tomwer_process_0"]["program"]),
                    SinoNormalizationTask.program_name(),
                )



# coding: utf-8
from __future__ import annotations

import functools
import logging

from processview.core.manager import DatasetState, ProcessManager
from processview.core.superviseprocess import SuperviseProcess
from silx.gui import qt

from tomwer.core.process.reconstruction.nabu.castvolume import CastVolumeTask
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.settings import get_lbsram_path, isOnLbsram
from tomwer.core.tomwer_object import TomwerObject
from tomwer.core.utils.lbsram import is_low_on_memory
from tomwer.core.volume.volumebase import TomwerVolumeBase
from tomwer.core.volume.volumefactory import VolumeFactory

from ..processingstack import FIFO, ProcessingThread

_logger = logging.getLogger(__name__)


class CastVolumeProcessStack(FIFO, qt.QObject):
    def __init__(self, process_id=None):
        qt.QObject.__init__(self)
        FIFO.__init__(self, process_id=process_id)
        self._dry_run = False

    def set_dry_run(self, dry_run):
        self._dry_run = dry_run

    def _process(self, data, configuration, callback=None):
        ProcessManager().notify_dataset_state(
            dataset=data,
            process=self,
            state=DatasetState.ON_GOING,
        )
        _logger.processStarted(
            f"start cast volume {str(data.get_identifier())} with parameters {configuration}"
        )

        if isinstance(data, TomwerVolumeBase):
            if data.data_url is not None:
                path = data.data_url.file_path()
            else:
                path = None
        elif isinstance(data, TomwerScanBase):
            path = data.path
        else:
            raise ValueError(
                f"data is expected to be an instance of {TomwerScanBase} or {TomwerVolumeBase}. {type(data)} provided"
            )

        if (
            path is not None
            and isOnLbsram(path)
            and is_low_on_memory(get_lbsram_path()) is True
        ):
            # if computer is running into low memory on lbsram skip it
            mess = f"low memory, skip volume cast {data}"
            ProcessManager().notify_dataset_state(
                dataset=data, process=self, state=DatasetState.SKIPPED
            )
            _logger.processSkipped(mess)
            if callback is not None:
                callback()
                self._end_computation(data=data, future_tomo_obj=None, callback=None)
        else:
            self._data_currently_computed = data
            try:
                self._computationThread.init(data=data, configuration=configuration)
            except ValueError as e:
                # initialization can fail (for example for cast volume is there is no volume or be case this will raise an error)
                # then we want to keep the thread active
                self._data_currently_computed = None
                ProcessManager().notify_dataset_state(
                    dataset=data, process=self, state=DatasetState.FAILED
                )
                _logger.processFailed(f"thread initialization failed. Error is {e}")
                if callback is not None:
                    callback()
                self._end_computation(data=data, future_tomo_obj=None, callback=None)
            else:
                # need to manage connect before starting it because
                fct_callback = functools.partial(
                    self._end_threaded_computation, callback
                )
                self._computationThread.finished.connect(fct_callback)
                self._computationThread.start()

    def _end_computation(self, data, future_tomo_obj, callback):
        """
        callback when the computation thread is finished

        :param scan: pass if no call to '_computationThread is made'
        """
        assert isinstance(data, TomwerObject)
        FIFO._end_computation(
            self, data=data, future_tomo_obj=future_tomo_obj, callback=callback
        )

    def _end_threaded_computation(self, callback=None):
        assert self._data_currently_computed is not None
        self._computationThread.finished.disconnect()
        if callback:
            callback()
        FIFO._end_threaded_computation(self)

    def _create_processing_thread(self, process_id=None) -> qt.QThread:
        return _ProcessingThread(process_id=process_id)


class _ProcessingThread(ProcessingThread, SuperviseProcess):
    """
    Thread use to execute the processing of the axis position
    """

    def __init__(self, process_id=None):
        SuperviseProcess.__init__(self, process_id=process_id)
        try:
            ProcessingThread.__init__(self, process_id=process_id)
        except TypeError:
            ProcessingThread.__init__(self)
        self._dry_run = False
        self._scan = None
        self._future_tomo_obj = None
        self._volume = None
        self._configuration = None

    @property
    def future_tomo_obj(self):
        return self._future_tomo_obj

    def set_dry_run(self, dry_run):
        self._dry_run = dry_run

    def init(self, data, configuration):
        if isinstance(data, TomwerVolumeBase):
            self._scan = None
            self._volume = data
        else:
            self._scan = data
            if len(data.latest_vol_reconstructions) < 1:
                raise ValueError(
                    "No reconstructed volume found. did you run the 'nabu volume reconstruction' process ?"
                )
            volume_identifier = data.latest_vol_reconstructions[0]
            self._volume = VolumeFactory.create_tomo_object_from_identifier(
                volume_identifier
            )
        self._data = data
        self._configuration = configuration

    def run(self):
        self.sigComputationStarted.emit()
        cast_volume_task = CastVolumeTask(
            process_id=self.process_id,
            varinfo=None,
            inputs={
                "volume": self._volume,
                "configuration": self._configuration,
                "scan": self._scan,
            },
        )
        try:
            cast_volume_task.run()
        except Exception as e:
            _logger.error(str(e))
            if self._scan is not None:
                # if scan is provided update status because otherwise only the volume state is updated
                ProcessManager().notify_dataset_state(
                    dataset=self._scan,
                    process=self,
                    state=DatasetState.FAILED,
                )
        else:
            if self._scan is not None:
                self._scan.cast_volume = cast_volume_task.outputs.volume
            if cast_volume_task.outputs.future_tomo_obj is not None:
                self._future_tomo_obj = cast_volume_task.outputs.future_tomo_obj



import pytest

from tomwer.gui.stitching.config.axisparams import StitcherAxisParams
from tomwer.tests.conftest import qtapp  # noqa F401


def test_StitcherAxisParams(
    qtapp,  # noqa F811
):
    """test the StitcherAxisParams widget"""
    with pytest.raises(TypeError):
        widget = StitcherAxisParams(axis="toto")
    widget = StitcherAxisParams(axis=0)
    assert widget.getConfiguration() == {
        "stitching": {
            "axis_0_params": "img_reg_method=nabu-fft;window_size=400",
        }
    }
    new_config = {
        "stitching": {
            "axis_0_params": "img_reg_method=skimage;window_size=50",
        },
    }
    widget.setConfiguration(new_config)
    assert widget.getConfiguration() == new_config



import gc
import os
import pickle
import shutil
import tempfile

import numpy
from orangecanvas.scheme.readwrite import literal_dumps
from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt

from orangecontrib.tomwer.widgets.control.VolumeSelector import VolumeSelectorOW
from tomwer.core.volume.hdf5volume import HDF5Volume


class TestVolumeSelectorOW(TestCaseQt):
    def setUp(self):
        super().setUp()
        self.window = VolumeSelectorOW()
        self.folder = tempfile.mkdtemp()
        self.volume_1 = HDF5Volume(
            file_path=os.path.join(self.folder, "vol_file.hdf5"),
            data_path="entry0000",
            data=numpy.linspace(0, 10, 100 * 100 * 3).reshape(3, 100, 100),
        )
        self.volume_1.save()
        self.volume_2 = HDF5Volume(
            file_path=os.path.join(self.folder, "vol_file2.hdf5"),
            data_path="entry0001",
            data=numpy.linspace(60, 120, 100 * 100 * 5).reshape(5, 100, 100),
        )
        self.volume_2.save()

    def tearDown(self):
        self.window.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.window.close()
        self.window = None
        shutil.rmtree(self.folder)
        gc.collect()

    def test(self):
        self.window.show()
        # check adding one volume
        self.window.addVolume(self.volume_1)
        assert self.window.widget.dataList.n_data() == 1
        # check behavior adding the same volume
        self.window.addVolume(self.volume_1)
        assert self.window.widget.dataList.n_data() == 1
        # check adding another volume
        self.window.addVolume(self.volume_2)
        assert self.window.widget.dataList.n_data() == 2
        # check removing volume(s)
        self.window.removeVolume(self.volume_1)
        assert self.window.widget.dataList.n_data() == 1
        self.window.removeVolume(self.volume_1)
        assert self.window.widget.dataList.n_data() == 1
        self.window.removeVolume(self.volume_2)
        assert self.window.widget.dataList.n_data() == 0

        # and adding them back
        self.window.addVolume(self.volume_1)
        self.window.addVolume(self.volume_2)
        assert self.window.widget.dataList.n_data() == 2

        # test serialization
        self.window._updateSettings()
        assert len(self.window._scanIDs) == 2
        pickle.dumps(self.window._scanIDs)
        literal_dumps(self.window._scanIDs)



import gc
import logging
import os
import pickle
import shutil
import tempfile
import uuid

import h5py
import numpy
from orangecanvas.scheme.readwrite import literal_dumps
from processview.core.manager import DatasetState, ProcessManager
from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt
from silx.io.url import DataUrl

from orangecontrib.tomwer.widgets.reconstruction.SADeltaBetaOW import (
    SADeltaBetaOW as _SADeltaBetaOW,
)
from tomwer.core import settings
from tomwer.core.utils.lbsram import mock_low_memory
from tomwer.core.process.reconstruction.scores import ComputedScore
from tomwer.core.utils.scanutils import MockNXtomo

logger = logging.getLogger(__name__)


class SADeltaBetaOW(_SADeltaBetaOW):
    def __init__(self, parent=None):
        self._scans_finished = []
        super().__init__(parent)

    def processing_finished(self, scan):
        # TODO: add message processing finished
        self._scans_finished.append(scan)

    @property
    def scans_finished(self):
        return self._scans_finished

    def close(self):
        self._scans_finished = {}
        super().close()


class TestProcessing(TestCaseQt):
    DIM = 100

    def setUp(self):
        super().setUp()
        self._source_dir = tempfile.mkdtemp()

        def create_scan(folder_name):
            _dir = os.path.join(self._source_dir, folder_name)
            return MockNXtomo(
                scan_path=_dir,
                n_ini_proj=20,
                n_proj=20,
                n_alignement_proj=2,
                create_final_flat=False,
                create_ini_dark=True,
                create_ini_flat=True,
                n_refs=1,
                dim=self.DIM,
            ).scan

        # create scans
        self.scan_1 = create_scan("scan_1")
        self.scan_2 = create_scan("scan_2")
        self.scan_3 = create_scan("scan_3")
        self._process_manager = ProcessManager()

        self.widget = SADeltaBetaOW()
        self.widget.show()

        def patch_score(cor):
            data = numpy.random.random(TestProcessing.DIM * TestProcessing.DIM)
            data = data.reshape(TestProcessing.DIM, TestProcessing.DIM)
            slice_file_path = os.path.join(
                self._source_dir, str(uuid.uuid1()) + ".hdf5"
            )
            data_url = DataUrl(
                file_path=slice_file_path, data_path="data", scheme="silx"
            )
            with h5py.File(slice_file_path, mode="a") as h5f:
                h5f["data"] = data
            return data_url, ComputedScore(
                tv=numpy.random.random(),
                std=numpy.random.random(),
            )

        self.widget._widget._processing_stack.patch_processing(patch_score)

    def tearDown(self):
        mock_low_memory(False)
        settings.mock_lsbram(False)
        self.widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.widget.close()
        self.widget = None
        self.qapp.processEvents()
        shutil.rmtree(self._source_dir)
        gc.collect()

    def test_serializing(self):
        pickle.dumps(self.widget.getConfiguration())

    def test_literal_dumps(self):
        literal_dumps(self.widget.getConfiguration())

    def testAutoFocusUnlock(self):
        self.widget.lockAutofocus(False)

        def manual_processing():
            self.widget.compute()
            self.qapp.processEvents()
            self.widget.wait_processing(5000)
            self.qapp.processEvents()

        self.widget.process(self.scan_1)
        manual_processing()
        self.assertEqual(
            self._process_manager.get_dataset_state(
                dataset_id=self.scan_1.get_identifier(),
                process=self.widget,
            ),
            DatasetState.WAIT_USER_VALIDATION,
        )

        self.widget.process(self.scan_2)
        manual_processing()
        self.assertEqual(len(self.widget.scans_finished), 0)
        self.assertEqual(
            self._process_manager.get_dataset_state(
                dataset_id=self.scan_1.get_identifier(),
                process=self.widget,
            ),
            DatasetState.SKIPPED,
        )
        self.assertEqual(
            self._process_manager.get_dataset_state(
                dataset_id=self.scan_2.get_identifier(),
                process=self.widget,
            ),
            DatasetState.WAIT_USER_VALIDATION,
        )
        self.widget.process(self.scan_3)
        manual_processing()
        self.widget.validateCurrentScan()
        self.assertEqual(
            self._process_manager.get_dataset_state(
                dataset_id=self.scan_3.get_identifier(),
                process=self.widget,
            ),
            DatasetState.SUCCEED,
        )
        # insure a cor has been registered
        self.assertNotEqual(
            self.scan_3.sa_delta_beta_params.selected_delta_beta_value, None
        )

    def testTestLbsram(self):
        mock_low_memory(True)
        settings.mock_lsbram(True)
        for scan in (self.scan_1, self.scan_2, self.scan_3):
            self.widget.process(scan)
            self.widget.wait_processing(5000)
            self.qapp.processEvents()

        for scan in (self.scan_1, self.scan_2, self.scan_3):
            with self.subTest(scan=str(scan)):
                self.assertEqual(
                    self._process_manager.get_dataset_state(
                        dataset_id=scan.get_identifier(),
                        process=self.widget,
                    ),
                    DatasetState.SKIPPED,
                )

    def testAutoFocusLock(self):
        self.widget.lockAutofocus(True)
        for scan in (self.scan_1, self.scan_2, self.scan_3):
            self.widget.process(scan)
            self.widget.wait_processing(5000)
            self.qapp.processEvents()

        for scan in (self.scan_1, self.scan_2, self.scan_3):
            with self.subTest(scan=str(scan)):
                self.assertEqual(
                    self._process_manager.get_dataset_state(
                        dataset_id=scan.get_identifier(),
                        process=self.widget,
                    ),
                    DatasetState.SUCCEED,
                )



import os
from tomoscan.series import Series
import pickle

from ewoksorange.bindings.qtapp import QtEvent

from orangecanvas.scheme.readwrite import literal_dumps

from tomwer.tests.conftest import qtapp  # noqa F401
from tomwer.core.utils.scanutils import MockNXtomo

from orangecontrib.tomwer.widgets.control.NXtomoConcatenate import NXtomoConcatenateOW


def test_NabuHelicalPrepareWeightsDoubleOW(
    qtapp,  # noqa F811
    tmp_path,
):
    """simple test of the _DeltaBetaSelectorDialog"""
    widget = NXtomoConcatenateOW()
    widget._loadSettings()
    # test settings serialization
    pickle.dumps(widget._ewoks_default_inputs)
    literal_dumps(widget._ewoks_default_inputs)

    scan_1 = MockNXtomo(
        scan_path=os.path.join(tmp_path, "scan1"),
        n_proj=10,
        n_ini_proj=10,
        scan_range=180,
        dim=20,
        energy=12.3,
    ).scan
    scan_2 = MockNXtomo(
        scan_path=os.path.join(tmp_path, "scan2"),
        n_proj=10,
        n_ini_proj=10,
        scan_range=180,
        dim=20,
        energy=12.3,
    ).scan

    output_scan_file = os.path.join(tmp_path, "concatenation.nx")
    assert not os.path.exists(output_scan_file)

    widget.widget.setConfiguration(
        {
            "output_file": output_scan_file,
        }
    )

    # process task
    finished = QtEvent()

    def finished_callback():
        finished.set()

    executor = widget.task_executor_queue
    executor.sigComputationEnded.connect(finished_callback)
    widget._process_series(series=Series(iterable=[scan_1, scan_2]))

    # wait until processing is finished
    assert finished.wait(timeout=10)
    assert os.path.exists(output_scan_file)



from __future__ import annotations

import logging
import os
from collections import OrderedDict
from typing import Any
from nxtomomill import converter as nxtomomill_converter
from nxtomomill.io import generate_default_edf_config, generate_default_h5_config
from nxtomomill.io.config import TomoEDFConfig as EDFConfig
from nxtomomill.io.config import TomoHDF5Config as HDF5Config
from silx.gui import qt

from tomoscan.esrf.volume.utils import guess_volumes
from tomoscan.identifier import BaseIdentifier
from tomoscan.esrf.identifier.hdf5Identifier import NXtomoScanIdentifier

from tomwer.core.scan.blissscan import BlissScan
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.tomwer_object import TomwerObject
from tomwer.core.volume.volumefactory import VolumeFactory
from tomwer.io.utils.tomoobj import DEFAULT_SCHEME_TO_VOL
from tomwer.gui.control.actions import CFGFileActiveLabel, NXTomomillParamsAction
from tomwer.gui.dialog.QDataDialog import QDataDialog
from tomwer.gui.dialog.QVolumeDialog import QVolumeDialog
from tomwer.gui.utils.inputwidget import (
    EDFConfigFileSelector,
    HDF5ConfigFileSelector,
    NXTomomillOutputDirSelector,
)
from tomwer.gui.utils.qt_utils import block_signals
from tomwer.gui.control.tomoobjdisplaymode import DisplayMode
from tomwer.gui.control.actions import TomoObjDisplayModeToolButton

logger = logging.getLogger(__name__)


class _DataListDialog(qt.QDialog):
    """A simple list of dataset path.BlissHDF5DataListDialog

    .. warning: the widget won't check for scan validity and will only
        emit the path to folders to the next widgets

    :param parent: the parent widget
    """

    sigUpdated = qt.Signal()
    """signal emitted when the list is updated"""

    def __init__(self, parent=None):
        qt.QDialog.__init__(self, parent)
        self.setLayout(qt.QVBoxLayout())
        # add list
        self.datalist = self.createDataList()
        self.layout().addWidget(self.datalist)
        # add buttons
        self._buttons = qt.QDialogButtonBox(parent=self)
        self._addButton = qt.QPushButton("Add", parent=self)
        self._buttons.addButton(self._addButton, qt.QDialogButtonBox.ActionRole)
        self._rmButton = qt.QPushButton("Remove", parent=self)
        self._buttons.addButton(self._rmButton, qt.QDialogButtonBox.ActionRole)
        self._rmAllButton = qt.QPushButton("Remove all", parent=self)
        self._buttons.addButton(self._rmAllButton, qt.QDialogButtonBox.ActionRole)

        self._sendSelectedButton = qt.QPushButton("Send selected", parent=self)
        self._buttons.addButton(
            self._sendSelectedButton, qt.QDialogButtonBox.AcceptRole
        )
        self.layout().addWidget(self._buttons)

        # connect signal / slot
        self._addButton.clicked.connect(self._callbackAddPath)
        self._rmButton.clicked.connect(self._removeSelected)
        self._rmAllButton.clicked.connect(self._callbackRemoveAllFolders)

    def selectAll(self):
        return self.datalist.selectAll()

    def clear(self):
        self.datalist.clear()

    def add(self, scan) -> tuple:
        added_objs = self.datalist.add(scan)
        self.datalist.setMySelection(added_objs)
        self.sigUpdated.emit()
        return added_objs

    def remove(self, scan):
        self.datalist.remove(scan)
        self.sigUpdated.emit()

    def n_scan(self):
        return len(self.datalist._myitems)

    def _callbackAddPath(self):
        """ """
        self.sigUpdated.emit()

    def _removeSelected(self):
        """remove all selected items"""
        selected_items = self.datalist.selectedItems()
        tomwer_objs_to_remove = [item.data(qt.Qt.UserRole) for item in selected_items]
        for tomwer_obj in tomwer_objs_to_remove:
            self.remove(tomwer_obj)
        self.sigUpdated.emit()

    def _callbackRemoveAllFolders(self):
        self.datalist.selectAll()
        self._removeSelected()

    def createDataList(self):
        raise NotImplementedError("Base class")


class _NXtomomillConfigFileDialog(qt.QDialog):
    def __init__(
        self,
        parent,
        warning: str,
        callback_new_config_file: Any,
        ConfigFileSelectorClass: Any,
    ):
        super().__init__(parent)
        self.setLayout(qt.QFormLayout())

        self._warningLabel = qt.QLabel(warning, self)
        self._warningLabel.setWordWrap(False)
        font = self._warningLabel.font()
        font.setItalic(True)
        font.setPixelSize(12)
        self._warningLabel.setFont(font)
        self._warningLabel.setAlignment(qt.Qt.AlignLeft | qt.Qt.AlignVCenter)
        #: add left intedation + lower size + italic
        self.layout().addRow(self._warningLabel)

        # handle configuration file
        self._configurationWidget = ConfigFileSelectorClass(self, try_load_cfg=False)
        self.layout().addRow("configuration file", self._configurationWidget)

        # button to create a default configuration file
        self._cdcfWidget = qt.QWidget(self)  # widget for layout
        self._cdcfWidget.setLayout(qt.QHBoxLayout())
        self._cdcfSpacer = qt.QWidget(self._cdcfWidget)
        self._cdcfSpacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum)
        self._cdcfWidget.layout().addWidget(self._cdcfSpacer)
        self._createDefaultConfigFileB = qt.QPushButton(
            "create new default configuration file", self._cdcfWidget
        )
        self._cdcfWidget.layout().addWidget(self._createDefaultConfigFileB)
        self.layout().addRow(self._cdcfWidget)

        # handle configuration file
        self._nxTomomillOutputWidget = NXTomomillOutputDirSelector()
        self.layout().addRow("nexus file output dir", self._nxTomomillOutputWidget)

        # buttons
        types = qt.QDialogButtonBox.Ok
        self.__buttons = qt.QDialogButtonBox(parent=self)
        self.__buttons.setStandardButtons(types)
        self.layout().addWidget(self.__buttons)

        # connect signal / slot
        self.__buttons.accepted.connect(self.accept)
        if callback_new_config_file is not None:
            self._createDefaultConfigFileB.released.connect(callback_new_config_file)

        # expose API
        self.sigConfigFileChanged = self._configurationWidget.sigConfigFileChanged
        self.sigOutputdirChanged = self._nxTomomillOutputWidget.sigChanged

    def setNewconfigFileCreationCallback(self, callback):
        self._createDefaultConfigFile = callback

    def getCFGFilePath(self):
        return self._configurationWidget.getCFGFilePath()

    def setCFGFilePath(self, cfg_file):
        self._configurationWidget.setCFGFilePath(cfg_file)

    def getOutputFolder(self):
        return self._nxTomomillOutputWidget.getOutputFolder()

    def setOutputDialog(self, output_dir):
        self._nxTomomillOutputWidget.setOutputFolder(output_dir)

    def accept(self):
        self.hide()

    def _createConfigFileSelector(self):
        raise NotImplementedError


class BlissHDF5DataListDialog(_DataListDialog):
    """Dialog used to load .h5 files only (used for nxtomomillOW when we need to do a conversion from bliss.h5 to NXtomo)"""

    def __init__(self, parent):
        assert isinstance(parent, _RawDataListMainWindow)
        _DataListDialog.__init__(self, parent)
        self._sendSelectedButton.setText("Send selected")

    def createDataList(self):
        return BlissScanList(self)

    def _callbackAddPath(self):  # pragma: no cover
        """Open file dialog to select HDF5 files or directories, allowing multi-selection"""
        dialog = qt.QFileDialog(self)
        dialog.setFileMode(qt.QFileDialog.ExistingFiles)
        dialog.setNameFilters(
            [
                "HDF5 files (*.h5 *.hdf5)",
                "Nexus files (*.h5 *.hdf5 *.nx *.nxs *.nexus)",
                "Any files (*)",
            ]
        )

        # Set default directory if available
        if os.environ.get("TOMWER_DEFAULT_INPUT_DIR", None) and os.path.exists(
            os.environ["TOMWER_DEFAULT_INPUT_DIR"]
        ):
            dialog.setDirectory(os.environ["TOMWER_DEFAULT_INPUT_DIR"])
        elif dialog.directory() != os.getcwd() or str(dialog.directory()).startswith(
            "/data"
        ):
            # if the directory as already been set by the user. Avoid redefining it
            pass
        elif os.path.isdir("/data"):
            dialog.setDirectory("/data")

        if not dialog.exec_():
            dialog.close()
            return

        filesSelected = dialog.selectedFiles()
        added_scans = []
        for file_ in filesSelected:
            added_scans = self.add(file_)
            if added_scans is None:
                continue
            added_scans.extend(added_scans)
        super()._callbackAddPath()
        self.datalist.setMySelection(added_scans)


class EDFDataListDialog(_DataListDialog):
    """Dialog used to load EDF directories and files for conversion"""

    def __init__(self, parent):
        assert isinstance(parent, EDFDataListMainWindow)
        _DataListDialog.__init__(self, parent)
        self._sendSelectedButton.setText("Convert and send selected")

    def createDataList(self):
        return EDFDataList(self)

    def _callbackAddPath(self):  # pragma: no cover
        """Open file dialog to select EDF directories or files"""
        dialog = qt.QFileDialog(self)
        dialog.setFileMode(qt.QFileDialog.DirectoryOnly)

        # Set default directory if available
        if os.environ.get("TOMWER_DEFAULT_INPUT_DIR", None) and os.path.exists(
            os.environ["TOMWER_DEFAULT_INPUT_DIR"]
        ):
            dialog.setDirectory(os.environ["TOMWER_DEFAULT_INPUT_DIR"])
        elif os.path.isdir("/data"):
            dialog.setDirectory("/data")

        if not dialog.exec_():
            dialog.close()
            return

        filesSelected = dialog.selectedFiles()
        added_scans = []
        for file_ in filesSelected:
            added = self.add(file_)
            assert added is not None
            added_scans.extend(added)
        super()._callbackAddPath()
        self.datalist.setMySelection(added_scans)


class _RawDataListMainWindow(qt.QMainWindow):
    sigNXTomoCFGFileChanged = qt.Signal(str)
    """signal emitted when the configuration file change"""

    sigUpdated = qt.Signal()
    """signal emitted when the list of raw data to convert change"""

    def __init__(
        self, parent, DataListConstructor, ConfigClass, warning, ConfigFileSelectorClass
    ):
        super().__init__(parent)
        self._widget = DataListConstructor(self)
        self.__configConstructor = ConfigClass
        # rework BlissHDF5DataListDialog layout
        self._subWidget = qt.QWidget(self)
        self._subWidget.setLayout(qt.QVBoxLayout())
        self._subWidget.layout().addWidget(self._widget.datalist)
        self._subWidget.layout().addWidget(self._widget._buttons)
        self.setCentralWidget(self._subWidget)

        self._dialog = _NXtomomillConfigFileDialog(
            self,
            warning=warning,
            callback_new_config_file=self._createNewConfigFile,
            ConfigFileSelectorClass=ConfigFileSelectorClass,
        )
        self._dialog.setWindowTitle("Select nxtomomill configuration file")

        # add toolbar
        toolbar = qt.QToolBar(self)
        toolbar.setMovable(False)
        toolbar.setFloatable(False)
        self.addToolBar(qt.Qt.TopToolBarArea, toolbar)

        # add filtering
        self._parametersAction = NXTomomillParamsAction(toolbar)
        toolbar.addAction(self._parametersAction)
        self._parametersAction.triggered.connect(self._parametersTriggered)

        # add tomo obj display mode
        self._tomoObjdisplayAction = TomoObjDisplayModeToolButton(self)
        toolbar.addWidget(self._tomoObjdisplayAction)

        # toolbar spacer
        spacer = qt.QWidget(toolbar)
        spacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum)
        toolbar.addWidget(spacer)

        # add information if cfg file is activate or not
        self._cfgStatusLabel = CFGFileActiveLabel(toolbar)
        toolbar.addWidget(self._cfgStatusLabel)

        # expose API
        self.getCFGFilePath = self._dialog.getCFGFilePath
        self.setCFGFilePath = self._dialog.setCFGFilePath
        self._sendSelectedButton = self._widget._sendSelectedButton
        self.add = self._widget.add
        self.n_scan = self._widget.n_scan
        self.datalist = self._widget.datalist

        # connect signal / slot
        self._dialog.sigConfigFileChanged.connect(self._cfgFileChanged)
        self._dialog.sigOutputdirChanged.connect(self.sigUpdated)
        self._widget.sigUpdated.connect(self.sigUpdated)
        self._tomoObjdisplayAction.sigDisplayModeChanged.connect(self.setDisplayMode)

    def setDisplayMode(self, *args, **kwargs):
        self.datalist.setDisplayMode(*args, **kwargs)

    def getOutputFolder(self):
        return self._dialog.getOutputFolder()

    def setOutputFolder(self, output_dir):
        self._dialog.setOutputDialog(output_dir)

    def _parametersTriggered(self):
        self._dialog.show()
        self._dialog.raise_()

    def _cfgFileChanged(self):
        cfg_file = self._dialog.getCFGFilePath()
        if cfg_file in (None, "") or not os.path.exists(cfg_file):
            self._cfgStatusLabel.setInactive()
        elif os.path.exists(cfg_file):
            try:
                self._load_cfg_file(cfg_file=cfg_file)
            except Exception:
                self._cfgStatusLabel.setInactive()
            else:
                self._cfgStatusLabel.setActive()
        else:
            self._cfgStatusLabel.setInactive()
        if cfg_file is None:
            cfg_file = ""
        self.sigNXTomoCFGFileChanged.emit(cfg_file)

    def _load_cfg_file(self, cfg_file):
        """The idea is to load the file before using it. This way the user can see if something is wrong with the file"""
        raise NotImplementedError("Base class")

    def getConfigInstance(self):
        """Return default HDF5Config or the one created from use input"""
        cfg_file = self._dialog.getCFGFilePath()
        if cfg_file in (None, ""):
            return self.__configConstructor()
        else:
            try:
                config = self.__configConstructor.from_cfg_file(cfg_file)
            except Exception:
                return self.__configConstructor()
            else:
                return config

    def createNXtomomillConfigFileDialog(self):  # pragma: no cover
        file_dialog = qt.QFileDialog()
        file_dialog.setAcceptMode(qt.QFileDialog.AcceptSave)
        file_dialog.setWindowTitle("Select file path to save configuration file")
        file_dialog.setNameFilters(
            [
                "Any file (*)",
                "Configuration file (*.txt *.cfg *.conf *.config)",
            ]
        )
        file_dialog.setFileMode(qt.QFileDialog.AnyFile)
        return file_dialog

    def _createNewConfigFile(self):
        """
        callback for the configuration file creation
        """
        raise NotImplementedError("Base class")


class BlissHDF5DataListMainWindow(_RawDataListMainWindow):
    def __init__(self, parent, *args, **kwargs):
        super().__init__(
            parent,
            DataListConstructor=BlissHDF5DataListDialog,
            ConfigClass=HDF5Config,
            warning="""Note: some parameters of the configuration file will be ignored:\n
            - input_file will be replaced from the item of the data list treated
            - output_file will be deduce from the output folder you defined""",
            ConfigFileSelectorClass=HDF5ConfigFileSelector,
            *args,
            **kwargs,
        )

    def _createNewConfigFile(self):  # pragma: no cover
        file_dialog = self.createNXtomomillConfigFileDialog()
        if file_dialog.exec_():
            files_selected = file_dialog.selectedFiles()
            if len(files_selected) > 0:
                file_path = files_selected[0]
                try:
                    # nxtomomill 1.1.0a5 or higher
                    configuration = (
                        generate_default_h5_config()  # pylint: disable=E1123
                    )

                except TypeError:
                    configuration = generate_default_h5_config(  # pylint: disable=E1123
                        config_3dxrd=False
                    )
                HDF5Config.dict_to_cfg(file_path=file_path, dict_=configuration)
                # if we create a configuration file then let consider we want to use it
                self.setCFGFilePath(file_path)

    def _load_cfg_file(self, cfg_file):
        HDF5Config.from_cfg_file(cfg_file)

    def setConfiguration(self, config: dict):
        if not isinstance(config, HDF5Config):
            raise TypeError
        self.datalist.configuration = config


class EDFDataListMainWindow(_RawDataListMainWindow):
    def __init__(self, parent, *args, **kwargs):
        super().__init__(
            parent,
            DataListConstructor=EDFDataListDialog,
            ConfigClass=EDFConfig,
            warning="""Note: some parameters of the configuration file will be ignored:\n
            - input_folder will be replaced from the item of the data list treated
            - dataset_basename (if will be deduce when the input folder is provided)
            - output_file will be deduce from the output folder you defined""",
            ConfigFileSelectorClass=EDFConfigFileSelector,
            *args,
            **kwargs,
        )

    def _createNewConfigFile(self):  # pragma: no cover
        file_dialog = self.createNXtomomillConfigFileDialog()
        if file_dialog.exec_():
            files_selected = file_dialog.selectedFiles()
            if len(files_selected) > 0:
                file_path = files_selected[0]
                configuration = generate_default_edf_config(level="advanced")
                EDFConfig.dict_to_cfg(file_path=file_path, dict_=configuration)
                # if we create a configuration file then let consider we want to use it
                self.setCFGFilePath(file_path)

    def _load_cfg_file(self, cfg_file):
        EDFConfig.from_cfg_file(cfg_file)


class GenericScanListDialog(_DataListDialog):
    """Dialog used to load EDFScan or HDF5 scans"""

    def createDataList(self):
        return GenericScanList(self)

    def _callbackAddPath(self):  # pragma: no cover
        """ """
        dialog = QDataDialog(self, multiSelection=True)
        dialog.setNameFilters(
            [
                "HDF5 files (*.h5 *.hdf5 *.nx *.nxs *.nexus)",
                "Nexus files (*.nx *.nxs *.nexus)",
                "Any files (*)",
            ]
        )

        if not dialog.exec_():
            dialog.close()
            return

        files_or_folders = dialog.files_selected()
        added_scans = []
        for file_or_folder in files_or_folders:
            new_scans = self.add(file_or_folder)
            if new_scans is not None:
                added_scans.extend(new_scans)
        super()._callbackAddPath()
        self.datalist.setMySelection(added_scans)


class GenericScanListWindow(qt.QMainWindow):
    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        self.setWindowFlags(qt.Qt.Widget)

        self._widget = GenericScanListDialog()
        self.datalist = self._widget.datalist
        self.setCentralWidget(self._widget)

        # toolbar
        toolbar = qt.QToolBar()
        toolbar.setMovable(False)
        toolbar.setFloatable(False)
        self.addToolBar(qt.Qt.TopToolBarArea, toolbar)

        tomoObjdisplayAction = TomoObjDisplayModeToolButton(self)
        toolbar.addWidget(tomoObjdisplayAction)

        # set up
        self.setDisplayMode(DisplayMode.SHORT)

        # expose API
        self.sigUpdated = self._widget.sigUpdated

        # connect signal / slot
        tomoObjdisplayAction.sigDisplayModeChanged.connect(self.setDisplayMode)

    def setDisplayMode(self, display_mode: DisplayMode):
        self.datalist.setDisplayMode(display_mode)

    # expose API
    def n_scan(self):
        return self._widget.n_scan()

    def add(self, *args, **kwargs):
        return self._widget.add(*args, **kwargs)


class VolumeListDialog(_DataListDialog):
    """Dialog used to load EDFScan or HEDF5 scans"""

    def createDataList(self):
        return VolumeList(self)

    def _callbackAddPath(self):  # pragma: no cover
        """ """
        dialog = QVolumeDialog(self)

        if not dialog.exec_():
            dialog.close()
            return
        files_or_folders = dialog.files_selected()
        added_volumes = []
        for file_or_folder in files_or_folders:
            new_volumes = self.add(file_or_folder)
            if new_volumes is not None:
                added_volumes.extend(new_volumes)
        super()._callbackAddPath()
        self.datalist.setMySelection(added_volumes)


class _TomwerObjectList(qt.QTableWidget):
    HEADER_NAMES = ("undefined object",)

    dataReady = qt.Signal(TomwerObject)

    listChanged = qt.Signal()
    """emit when containt of the list changed"""

    def __init__(self, parent):
        self._copy_target = None
        qt.QTableWidget.__init__(self, parent)
        self.setRowCount(0)
        self.setColumnCount(1)
        self.setHorizontalHeaderLabels(self.HEADER_NAMES)
        self.setSortingEnabled(True)
        self.verticalHeader().hide()
        if hasattr(self.horizontalHeader(), "setSectionResizeMode"):  # Qt5
            self.horizontalHeader().setSectionResizeMode(0, qt.QHeaderView.Stretch)
        else:  # Qt4
            self.horizontalHeader().setResizeMode(0, qt.QHeaderView.Stretch)
        self.setAcceptDrops(True)

        self._myitems = OrderedDict()
        # key is the TomoObject identifier and value is the QTableWidgetItem.
        # Text is the identifier, QTableWidgetItem data under UserRole is the TomoObject object

        # QMenu
        self.menu = qt.QMenu(self)
        self._copyAction = qt.QAction("copy")
        self._copyAction.triggered.connect(self._copyRequested)
        self._pasteAction = qt.QAction("paste")
        self._pasteAction.triggered.connect(self._pasteRequested)
        self.menu.addAction(self._copyAction)
        self.menu.addAction(self._pasteAction)

        self.setDisplayMode(DisplayMode.SHORT)

    def setDisplayMode(self, mode: DisplayMode) -> None:
        self._displayMode = DisplayMode.from_value(mode)
        self._update()

    def n_data(self):
        return len(self._myitems)

    def remove_item(self, item):
        """Remove a given folder"""
        try:
            del self._myitems[item.data(qt.Qt.UserRole).get_identifier().to_str()]
        except RuntimeError:
            # look like this could failed on some Qt or PyQt version (see issue 802)
            # no much that we can do so ignore it.
            pass
        self._update()

    def _sendSelected(self):
        raise NotImplementedError
        for _, item in self._myitems.items():
            data = item.data(qt.Qt.UserRole)
            self.dataReady.emit(data)

    def remove(self, data: TomwerObject | str | None):
        if data is None:
            return
        if isinstance(data, str):
            data = self._getTomoObject(data, allow_several=False)
            if data is None:
                logger.warning(f"unable to get a {TomwerObject} from {data}")
                return

            identifier_as_str = data.get_identifier().to_str()
        if not isinstance(data, TomwerObject):
            raise ValueError(f"{data} is not a TomwerObject")

        identifier_as_str = data.get_identifier().to_str()

        if identifier_as_str not in self._myitems:
            logger.info(f"{identifier_as_str} not in {self._myitems}")
            return
        else:
            item = self._myitems[identifier_as_str]
            self.remove_item(item)
            self.listChanged.emit()

    def _update(self):
        tomwer_objects = [
            self._myitems[identifier_as_str].data(qt.Qt.UserRole)
            for identifier_as_str in self._myitems.keys()
        ]
        self.clear()
        with block_signals(self):
            for tomwer_object in tomwer_objects:
                self.add(tomwer_object)
        self.sortByColumn(0, self.horizontalHeader().sortIndicatorOrder())
        self.listChanged.emit()

    def _getTomoObject(self, obj: str, allow_several: bool = False):
        """
        some rules to return a TomwerObject from an object (probably a path) from children class

        :param allow_several: allow returning several objects for one input. This can be the case for example if 'obj' is a file path.
        """
        raise NotImplementedError("Base class")

    def add(self, obj: TomwerObject | BaseIdentifier | str | None) -> tuple | None:
        """add a data"""
        if obj is None:
            return
        if isinstance(obj, TomwerObject):
            # remove heavy cache obj. This could bring troubles if some processing is done.
            # but I guess this is better than keeping in memory volumes...
            # nevertheless processing using volume.data should keep a reference on the data.
            obj._clear_heavy_cache()
        elif isinstance(obj, str):
            obj = self._getTomoObject(obj, allow_several=True)
            if obj is None:
                return
            elif isinstance(obj, (tuple, list)):
                # in the case it contains in fact several objects. Related to allow_several=True
                new_objs = []
                for o in obj:
                    new_objs.extend(list(self.add(o)))
                return new_objs
            elif not isinstance(obj, TomwerObject):
                raise TypeError(
                    f"return object from _getTomoObject is not a {TomwerObject} but {type(obj)}"
                )
        elif isinstance(obj, BaseIdentifier):
            try:
                obj = ScanFactory.create_tomo_object_from_identifier(obj)
            except Exception:
                try:
                    obj = VolumeFactory.create_tomo_object_from_identifier(obj)
                except Exception:
                    raise ValueError(f"Unable to create an TomwerObject from {obj}")

        else:
            raise TypeError(
                f"is expected to be an instance of str or of {TomwerObject} but {type(obj)}"
            )

        identifier_as_str = obj.get_identifier().to_str()
        if identifier_as_str in self._myitems:
            # in this case we will update the value. Because if the same identifier already exists at launch time
            # it will create two different objects that could be an issue.
            # this is better to have it unified
            with block_signals(self):
                self.remove(self._myitems[identifier_as_str].data(qt.Qt.UserRole))

        item = qt.QTableWidgetItem()
        item.setText(self.getTextToDisplay(obj))
        item.setFlags(qt.Qt.ItemIsEnabled | qt.Qt.ItemIsSelectable)
        item.setData(qt.Qt.UserRole, obj)

        row = self.rowCount()
        self.setRowCount(row + 1)
        self.setItem(row, 0, item)
        self._myitems[identifier_as_str] = item
        self.listChanged.emit()
        return (obj,)

    def getTextToDisplay(self, obj: TomwerObject):
        if self._displayMode is DisplayMode.SHORT:
            return obj.get_identifier().short_description()
        elif self._displayMode is DisplayMode.URL:
            return obj.get_identifier().to_str()
        else:
            raise ValueError(
                f"Requested display mode {self._displayMode} is not handled"
            )

    def setMySelection(self, datasets: tuple[str | TomwerObject]):
        if datasets is None:
            datasets = tuple()

        def convert_dataset(dataset: str | TomwerObject):
            if isinstance(dataset, TomwerObject):
                return dataset.get_identifier().to_str()
            else:
                return dataset

        datasets = tuple([convert_dataset(dataset) for dataset in datasets])
        for row in range(self.rowCount()):
            item = self.item(row, 0)
            select_it = (
                item.data(qt.Qt.UserRole).get_identifier().to_str() in datasets
                or item.text() in datasets
            )
            item.setSelected(select_it)

    def clear(self):
        """Remove all items on the list"""
        self._myitems = OrderedDict()
        qt.QTableWidget.clear(self)
        self.setRowCount(0)
        self.setHorizontalHeaderLabels(self.HEADER_NAMES)
        if hasattr(self.horizontalHeader(), "setSectionResizeMode"):  # Qt5
            self.horizontalHeader().setSectionResizeMode(0, qt.QHeaderView.Stretch)
        else:  # Qt4
            self.horizontalHeader().setResizeMode(0, qt.QHeaderView.Stretch)

    def dropEvent(self, event):
        if event.mimeData().hasFormat("text/uri-list"):
            added_scans = set()
            with block_signals(self):
                for url in event.mimeData().urls():
                    new_scans = self.add(str(url.path()))
                    if new_scans is not None:
                        for new_scan in new_scans:
                            if not isinstance(new_scan, TomwerObject):
                                raise ValueError(
                                    f"new_scan should be an instance of {TomwerObject} and not {type(new_scan)}"
                                )
                        added_scans.update(new_scans)
            self.setMySelection(added_scans)
            self.listChanged.emit()

    def supportedDropActions(self):
        """Inherited method to redefine supported drop actions."""
        return qt.Qt.CopyAction | qt.Qt.MoveAction

    def dragEnterEvent(self, event):
        if hasattr(event, "mimeData") and event.mimeData().hasFormat("text/uri-list"):
            event.accept()
            event.setDropAction(qt.Qt.CopyAction)
        else:
            try:
                qt.QListWidget.dragEnterEvent(self, event)
            except TypeError:
                pass

    def dragMoveEvent(self, event):
        if hasattr(event, "mimeDatamyitems") and event.mimeDatamyitems().hasFormat(
            "text/uri-list"
        ):
            event.setDropAction(qt.Qt.CopyAction)
            event.accept()
        else:
            try:
                qt.QListWidget.dragMoveEvent(self, event)
            except TypeError:
                pass

    def _datasetAt(self, point):
        item = self.itemAt(point)
        if item is not None:
            return item.data(qt.Qt.UserRole)

    def contextMenuEvent(self, event):
        self._copy_target = self._datasetAt(event.pos())
        self._copyAction.setVisible(self._copy_target is not None)
        self.menu.exec_(event.globalPos())

    def _copyRequested(self):
        clipboard = qt.QGuiApplication.clipboard()

        def get_info(item):
            user_data = item.data(qt.Qt.UserRole)
            if isinstance(user_data, TomwerObject):
                return user_data.get_identifier().to_str()
            else:
                return str(user_data)

        selection = [get_info(item) for item in self.selectedItems()]
        clipboard.setText("\n".join(selection))

    def _pasteRequested(self):
        clipboard = qt.QGuiApplication.clipboard()
        identifiers = clipboard.text()
        # handle paste of several lines...
        identifiers.replace(";", "\n")
        for identifier in identifiers.split("\n"):
            try:
                self.add(identifier)
            except Exception as e:
                logger.error(f"Failed to add '{identifier}'. Error is {e}")


class GenericScanList(_TomwerObjectList):
    """Data list able to manage directories (EDF/HDF5?) or files (HDF5)"""

    HEADER_NAMES = ("dataset",)

    def _getTomoObject(self, obj: str, allow_several: bool = False):
        """
        some rules to return a TomwerObject from an object (probably a path) from children class
        """
        return self.getScanObject(obj, allow_several=allow_several)

    @staticmethod
    def getScanObject(obj, allow_several: bool = False):
        if not isinstance(obj, str):
            raise TypeError(f"obj is an instance of {type(obj)} when {str} expected")
        try:
            scan_obj = ScanFactory.create_tomo_object_from_identifier(obj)
        except ValueError as e1:
            try:
                if allow_several:
                    scan_obj = ScanFactory.create_scan_objects(obj)
                else:
                    scan_obj = ScanFactory.create_scan_object(obj)
            except Exception as e2:
                logger.warning(
                    f"Unable to create scan object from identifier ({e1}) or as a file or a folder ({e2})"
                )
                return None
            else:
                return scan_obj
        else:
            return scan_obj


class BlissScanList(_TomwerObjectList):
    """
    Widget dedicated to convert bliss entries to TomwerScanBase)
    """

    HEADER_NAMES = ("bliss scan",)

    def __init__(self, parent):
        self._configuration = HDF5Config()
        super().__init__(parent)

    @property
    def configuration(self):
        return self._configuration

    @configuration.setter
    def configuration(self, config: HDF5Config):
        assert isinstance(
            config, HDF5Config
        ), "config is expect to be an instance of dict"
        self._configuration = config

    def _update(self):
        list_data = list(self._myitems.keys())
        self.clear()
        for data in list_data:
            self.add(data)
        self.sortByColumn(0, self.horizontalHeader().sortIndicatorOrder())

    def add(self, data) -> tuple:
        """Add the path folder d in the scan list

        :param data: the path of the directory to add
        """
        possible_entries = []
        if os.path.exists(data):
            if not BlissScan.is_bliss_file(data):
                msg = qt.QMessageBox(self)
                msg.setIcon(qt.QMessageBox.Warning)
                types = qt.QMessageBox.Ok | qt.QMessageBox.Cancel
                msg.setStandardButtons(types)

                if NXtomoScan.is_nexus_nxtomo_file(data):
                    text = (
                        f"The input file `{data}` seems to contain `NXTomo` entries. "
                        "and no valid `Bliss` valid entry. \n"
                        "This is probably not a Bliss file. Do you still want to translate ?"
                    )
                else:
                    text = (
                        f"The input file `{data}` does not seems to contain any "
                        "valid `Bliss` entry. \n"
                        "This is probably not a Bliss file. Do you still want to translate ?"
                    )
                msg.setText(text)
                if msg.exec_() != qt.QMessageBox.Ok:
                    return

            try:
                for entry in nxtomomill_converter.get_bliss_tomo_entries(
                    data, self.configuration
                ):
                    possible_entries.append(entry)
            except Exception:
                logger.error(f"Faild to find entries for {data}")
                return
            else:
                file_path = data
        else:
            identifier = NXtomoScanIdentifier.from_str(data)
            possible_entries.append(identifier.data_path)
            file_path = identifier.file_path
        created_scans = []
        for entry in possible_entries:
            scan = NXtomoScan(scan=file_path, entry=entry)
            scan_objs = super().add(scan)
            if scan_objs is not None:
                for scan_obj in scan_objs:
                    created_scans.append(scan_obj)
        return created_scans


class EDFDataList(_TomwerObjectList):
    HEADER_NAMES = ("edf scan url",)

    def __init__(self, parent):
        super().__init__(parent)

    def _getTomoObject(self, obj: str, allow_several: bool = False):
        # note: allow several is not used there. No sense for EDF scan
        try:
            tomo_obj = ScanFactory.create_scan_object(obj)
        except Exception as e1:
            try:
                tomo_obj = ScanFactory.create_tomo_object_from_identifier(obj)
            except Exception as e2:
                logger.warning(
                    f"Unable to create a EDFVolume from {obj}. Error is {e1} or {e2}"
                )
                return

        if isinstance(tomo_obj, EDFTomoScan):
            return tomo_obj
        else:
            logger.warning(
                f"Unable to create a EDFVolume from {obj}. But creates a {type(tomo_obj)}. Atre you sure provided path leads to EDF acquisition ?"
            )
            return None

    def getEDFTomoScan(self, scan_id: str, default=None):
        if scan_id in self._myitems:
            return self._myitems[scan_id].data(qt.Qt.UserRole)
        else:
            return default


class VolumeList(_TomwerObjectList):
    """
    Widget dedicated to the VolumeBase object
    """

    HEADER_NAMES = ("volume url",)

    def _getTomoObject(self, obj: str, allow_several: bool = False):
        return self.getVolumeObject(obj=obj, allow_several=allow_several)

    @staticmethod
    def getVolumeObject(obj, allow_several: bool = False, warn=True):
        """
        get volume from identifier... even if not contained in the list of items (using factory)
        """
        try:
            tomo_obj = VolumeFactory.create_tomo_object_from_identifier(obj)
        except Exception as e:
            try:
                tomo_obj = guess_volumes(obj, scheme_to_vol=DEFAULT_SCHEME_TO_VOL)
            except Exception:
                if warn:
                    logger.warning(
                        f"Unable to create a volume from {obj}. Error is {e}"
                    )
                return None
            else:
                # filter potential 'nabu histogram'
                if tomo_obj is not None:

                    def is_not_histogram(vol_identifier):
                        return not (
                            hasattr(vol_identifier, "data_path")
                            and vol_identifier.data_path.endswith("histogram")
                        )

                    tomo_obj = tuple(filter(is_not_histogram, tomo_obj))

                if tomo_obj is None or len(tomo_obj) == 0:
                    logger.warning(f"Unable to create a volume from {obj}.")
                    return None
                else:
                    if len(tomo_obj) > 1 and not allow_several:
                        if warn:
                            logger.warning(
                                f"more than one volume deduce from {obj}. Will only take the first one ({tomo_obj[0]})"
                            )
                        return tomo_obj[0]
                    elif len(tomo_obj) == 1:
                        return tomo_obj[0]
                    else:
                        return tomo_obj
        else:
            return tomo_obj

    def getVolume(self, volume_id: str, default=None):
        """
        get volume from id contained in the current items, else default
        """
        if volume_id in self._myitems:
            return self._myitems[volume_id].data(qt.Qt.UserRole)
        else:
            return default


class TomoObjList(_TomwerObjectList):
    HEADER_NAMES = ("tomo object",)

    def _getTomoObject(self, obj: str, allow_several: bool = False):
        try:
            tomo_obj = VolumeList.getVolumeObject(obj=obj, allow_several=allow_several)
        except Exception:
            pass
        else:
            if tomo_obj not in (None, tuple()):
                return tomo_obj

        try:
            tomo_obj = GenericScanList.getScanObject(
                obj=obj, allow_several=allow_several
            )
        except Exception:
            return None
        else:
            return tomo_obj

    def getTextToDisplay(self, obj: TomwerObject):
        return obj.get_identifier().short_description()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Application to stop possible running data listener (and free used port)
"""

import argparse
import sys
from signal import SIGKILL, SIGTERM

from tomwer.core.process.control.datalistener.rpcserver import (
    send_signal_to_local_rpc_servers,
)
from tomwer.core.settings import JSON_RPC_PORT


def getinputinfo():
    return "tomwer stop-data-listener [[--port --sigkill]]"


def main(argv):
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "--port",
        default=JSON_RPC_PORT,
        help="Define the port occupy by the rpc-server",
    )
    parser.add_argument(
        "--sigkill",
        action="store_true",
        default=False,
        help="send a SIGKILL signal instead of a SIGTERM",
    )
    options = parser.parse_args(argv[1:])
    signal = SIGTERM
    if options.sigkill:
        signal = SIGKILL
    send_signal_to_local_rpc_servers(port=int(options.port), signal=signal)


if __name__ == "__main__":
    main(sys.argv)



import numpy

from tomwer.gui.utils.lineselector.lineselector import QSliceSelectorDialog
from tomwer.tests.conftest import qtapp  # noqa F401


def test_line_selector(
    qtapp,  # noqa F811
):
    dialog = QSliceSelectorDialog(parent=None)
    dialog.setData(
        numpy.ones((100, 100)),
    )
    dialog.mainWidget.addSlice(2)
    dialog.mainWidget.addSlice(12)
    assert dialog.getSelection() == (2, 12)
    dialog.mainWidget.removeSlice(12)
    assert dialog.getSelection() == (2,)
    dialog.setSelection((12, 23))
    assert dialog.getSelection() == (12, 23)
    assert dialog.mainWidget.nSelected() == 2



import gc
import logging
import os
import pickle
import shutil
import tempfile
import time
from glob import glob

import h5py
from orangecanvas.scheme.readwrite import literal_dumps
from silx.gui import qt
from silx.gui.utils.testutils import SignalListener, TestCaseQt
from nxtomo.nxobject.nxdetector import FOV

from orangecontrib.tomwer.widgets.reconstruction.NabuOW import NabuOW
from tomwer.core.process.reconstruction.nabu.utils import _NabuMode
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.settings import mock_lsbram
from tomwer.core.utils.lbsram import mock_low_memory
from tomwer.synctools.darkref import QDKRFRP
from tomwer.tests.datasets import TomwerCIDatasets

logging.disable(logging.INFO)


class TestNabuWidget(TestCaseQt):
    """class testing the NabuOW"""

    def setUp(self):
        TestCaseQt.setUp(self)
        self._recons_params = QDKRFRP()
        self.widget = NabuOW(parent=None)
        self.scan_dir = tempfile.mkdtemp()
        # create dataset
        self.master_file = os.path.join(self.scan_dir, "frm_edftomomill_twoentries.nx")
        shutil.copyfile(
            TomwerCIDatasets.get_dataset(
                "h5_datasets/frm_edftomomill_twoentries.nx",
            ),
            self.master_file,
        )
        self.scan = NXtomoScan(scan=self.master_file, entry="entry0000")
        # create listener for the nabu widget
        self.signal_listener = SignalListener()

        # connect signal / slot
        self.widget.sigScanReady.connect(self.signal_listener)

        # set up
        mock_low_memory(True)
        mock_lsbram(True)
        self.widget.setDryRun(dry_run=True)

    def tearDown(self):
        mock_low_memory(False)
        mock_lsbram(False)
        self.widget.sigScanReady.disconnect(self.signal_listener)
        self._recons_params = None
        self.scan = None
        self.widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.widget.close()
        self.widget = None
        gc.collect()

    def test_serializing(self):
        pickle.dumps(self.widget.getConfiguration())

    def test_literal_dumps(self):
        literal_dumps(self.widget.getConfiguration())

    def testLowMemory(self):
        """Make sure no reconstruction is started if we are low in memory in
        lbsram"""
        self.assertEqual(len(glob(os.path.join(self.scan_dir, "*.cfg"))), 0)
        self.widget.process(self.scan)
        self.wait_processing()
        self.assertEqual(len(glob(os.path.join(self.scan_dir, "*.cfg"))), 0)

    def wait_processing(self):
        timeout = 10
        while timeout >= 0 and self.signal_listener.callCount() == 0:
            timeout -= 0.1
            time.sleep(0.1)
        if timeout <= 0.0:
            raise TimeoutError("nabu widget never end processing")

    def patch_fov(self, value: str):
        with h5py.File(self.scan.master_file, mode="a") as h5s:
            for entry in ("entry0000", "entry0001"):
                entry_node = h5s[entry]
                if "instrument/detector/field_of_view" in entry_node:
                    del entry_node["instrument/detector/field_of_view"]
                entry_node["instrument/detector/field_of_view"] = value

    def testSetConfiguration(self):
        """Make sure the configuration evolve from scan information"""
        self.assertEqual(self.widget.getMode(), _NabuMode.FULL_FIELD)
        self.patch_fov(value=FOV.HALF.value)
        self.widget.process(self.scan)
        self.wait_processing()
        self.assertEqual(self.widget.getMode(), _NabuMode.HALF_ACQ)
        self.patch_fov(value=FOV.FULL.value)
        self.scan.clear_caches()
        self.widget.process(self.scan)
        self.wait_processing()
        self.assertEqual(self.widget.getMode(), _NabuMode.FULL_FIELD)



import gc
import pickle

from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt

from orangecontrib.tomwer.widgets.control.EmailOW import EmailOW
from orangecanvas.scheme.readwrite import literal_dumps


class TestEmailOW(TestCaseQt):
    def setUp(self):
        super().setUp()
        self.widget = EmailOW()

    def tearDown(self):
        self.widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.widget.close()
        self.widget = None
        gc.collect()

    def test(self):
        self.widget.show()

    def test_serializing(self):
        pickle.dumps(self.widget.getConfiguration())

    def test_literal_dumps(self):
        literal_dumps(self.widget.getConfiguration())



# coding: utf-8
from __future__ import annotations


import os
import tempfile
from glob import glob

import pytest
from ewokscore.graph import analysis, load_graph
from ewokscore.graph.validate import validate_graph

from tomwer.core.utils.scanutils import HDF5MockContext
from nabu.pipeline.config import get_default_nabu_config
from nabu.pipeline.fullfield.nabu_config import (
    nabu_config as nabu_fullfield_default_config,
)

try:
    from nabu.pipeline.fullfield.reconstruction import (  # noqa F401
        FullFieldReconstructor,
    )
except ImportError:
    try:
        from nabu.pipeline.fullfield.local_reconstruction import (  # noqa F401
            ChunkedReconstructor,
        )
    except ImportError:
        has_nabu = False
    else:
        has_nabu = True
else:
    has_nabu = True

pytest.mark.skipif(condition=not has_nabu, reason="nabu not installed")


classes_to_test = {
    "darkref": "tomwer.core.process.reconstruction.darkref.darkrefs.DarkRefs",
    "axis": "tomwer.core.process.reconstruction.axis.axis.AxisTask",
    "nabu slices": "tomwer.core.process.reconstruction.nabu.nabuslices.NabuSlices",
}


@pytest.mark.parametrize("node_name, node_qual_name", classes_to_test.items())
def test_single_class_instanciation(node_name, node_qual_name):
    with HDF5MockContext(
        scan_path=os.path.join(tempfile.mkdtemp(), "scan_test"), n_proj=100
    ) as scan:
        # insure no cfg yet
        assert len(glob(os.path.join(scan.path, "*.cfg"))) == 0

        graph = load_graph(
            {
                "nodes": [
                    {
                        "id": node_name,
                        "task_type": "class",
                        "task_identifier": node_qual_name,
                        "default_inputs": [
                            {
                                "name": "data",
                                "value": scan,
                            },
                            {
                                "name": "nabu_params",
                                "value": get_default_nabu_config(
                                    nabu_fullfield_default_config
                                ),
                            },
                        ],
                    },
                ]
            }
        )

        assert graph.is_cyclic is False, "graph is expected to be acyclic"
        assert analysis.start_nodes(graph.graph) == {
            node_name
        }, "graph is expected to have only on start nodes"
        validate_graph(graph.graph)
        result = graph.execute(varinfo=None)

        assert analysis.end_nodes(graph.graph) == {
            node_name
        }, "graph is expected to have only one end node"
        assert "data" in result, "data is expected to be part of the output_values"



from tomwer.tests.conftest import qtapp  # noqa F401
from tomwer.gui.reconstruction.nabu.helical import HelicalPrepareWeightsDouble


def test_DeltaBetaSelector(
    qtapp,  # noqa F811
):
    """simple test of the _DeltaBetaSelectorDialog"""
    widget = HelicalPrepareWeightsDouble()
    assert widget.getConfiguration() == {
        "processes_file": "{scan_parent_dir_basename}/{scan_dir_name}/map_and_doubleff.hdf5",
        "transition_width_vertical": 50.0,
        "transition_width_horizontal": 50.0,
    }
    config = {
        "processes_file": "test.hdf5",
        "transition_width_vertical": 12.5,
        "transition_width_horizontal": 11.5,
    }
    widget.setConfiguration(config=config)
    assert widget.getConfiguration() == config



# coding: utf-8
from __future__ import annotations

import copy
import logging

import tomoscan.esrf.scan.utils
from orangewidget import gui
from orangewidget.settings import Setting
from orangewidget.widget import Input, Output
from processview.core.manager import DatasetState, ProcessManager
from silx.gui import qt

import tomwer.core.process.reconstruction.darkref.darkrefscopy
from orangecontrib.tomwer.orange.managedprocess import SuperviseOW
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.gui.reconstruction.darkref.darkrefcopywidget import DarkRefAndCopyWidget
from tomwer.synctools.darkref import QDKRFRP
from tomwer.synctools.stacks.reconstruction.dkrefcopy import DarkRefCopyProcessStack
from tomwer.utils import docstring

from ..utils import WidgetLongProcessing

_logger = logging.getLogger(__name__)


class DarkRefAndCopyOW(SuperviseOW, WidgetLongProcessing):
    """
    A simple widget managing the copy of an incoming folder to an other one

    :param parent: the parent widget
    """

    # note of this widget should be the one registered on the documentation
    name = "reduced darks and flats"
    id = "orange.widgets.tomwer.darkrefs"
    description = (
        "This widget will generate reduced darks and flats for a received scan "
    )
    icon = "icons/darkref.svg"
    priority = 15
    keywords = ["tomography", "dark", "darks", "ref", "refs", "flat", "flats"]

    want_main_area = True
    resizing_enabled = True

    _ewoks_default_inputs = Setting({"data": None, "dark_ref_params": None})

    sigScanReady = qt.Signal(TomwerScanBase)
    """Signal emitted when a scan is ready"""

    ewokstaskclass = (
        tomwer.core.process.reconstruction.darkref.darkrefscopy.DarkRefsCopy
    )

    class Inputs:
        data = Input(
            name="data",
            type=TomwerScanBase,
            doc="one scan to be process",
            multiple=True,
        )
        reduced_darks = Input(
            name="reduced dark(s)",
            type=dict,
            doc="dict containing reduced dark(s)",
            multiple=False,
        )
        reduced_flats = Input(
            name="reduced flat(s)",
            type=dict,
            doc="dict of containing reduced flat(s)",
            multiple=False,
        )

    class Outputs:
        data = Output(name="data", type=TomwerScanBase, doc="one scan to be process")
        reduced_darks = Output(
            name="reduced dark(s)",
            type=dict,
            doc="DataUrl of the reduced dark. Key is the relative position, value is the reduced dark",
        )
        reduced_flats = Output(
            name="reduced flat(s)",
            type=dict,
            doc="DataUrl of the reduced flat. Key is the relative position, value is the reduced flat",
        )

    def __init__(self, parent=None, reconsparams: QDKRFRP | None = None):
        """

        :param reconsparams: reconstruction parameters
        """
        SuperviseOW.__init__(self, parent)
        WidgetLongProcessing.__init__(self)
        recons_params = reconsparams or QDKRFRP()
        self._processing_stack = DarkRefCopyProcessStack(process_id=self.process_id)

        dark_ref_params = self._ewoks_default_inputs.get("dark_ref_params", None)
        if dark_ref_params not in ({}, None):
            try:
                recons_params.dkrf.load_from_dict(dark_ref_params)
            except Exception:
                _logger.warning("fail to load reconstruction settings")

        self.widget = DarkRefAndCopyWidget(
            save_dir=self._processing_stack._save_dir,
            parent=self,
            reconsparams=recons_params,
            process_id=self.process_id,
        )
        self._layout = gui.vBox(self.mainArea, self.name).layout()
        self._layout.addWidget(self.widget)
        self.setForceSync = self.widget.setForceSync

        # expose API
        self.setModeAuto = self.widget.set_mode_auto
        self.setRefsFromScan = self.widget.setRefsFromScan
        self.setCopyActive = self.widget.setCopyActive

        # connect signal / slot
        self.widget.sigProcessingStart.connect(self._startProcessing)
        self.widget.sigProcessingEnd.connect(self._endProcessing)
        self.widget.sigScanReady.connect(self.signalReady)
        self.widget.recons_params.sigChanged.connect(self._updateSettingsVals)
        self.widget.sigModeAutoChanged.connect(self._updateSettingsVals)
        self.widget.sigCopyActivationChanged.connect(self._updateSettingsVals)
        self.widget.sigClearCache.connect(self._processing_stack.clear_cache)
        self._processing_stack.sigComputationStarted.connect(self._startProcessing)
        self._processing_stack.sigComputationEnded.connect(self._endProcessing)
        self._processing_stack.sigRefSetted.connect(self.widget.setRefSetBy)

        # load some other copy parameters
        if dark_ref_params not in ({}, None):
            try:
                if "activate" in dark_ref_params:
                    self.widget.setCopyActive(dark_ref_params.pop("activate"))
                if "auto" in dark_ref_params:
                    auto_mode = dark_ref_params.pop("auto")
                    # insure backward compatibility. Has beem saved as a tuple (this was a typo)
                    if isinstance(auto_mode, tuple) and isinstance(auto_mode[0], bool):
                        auto_mode = auto_mode[0]
                    self.widget.setModeAuto(auto_mode)
            except Exception:
                _logger.warning("fail to load reconstruction settings")

        self.setCaption(self.windowTitle())

    def __new__(cls, *args, **kwargs):
        # ensure backward compatibility with 'static_input'
        static_input = kwargs.get("stored_settings", {}).get("static_input", None)
        if static_input not in (None, {}):
            _logger.warning(
                "static_input has been deprecated. Will be replaced by _ewoks_default_inputs in the workflow file. Please save the workflow to apply modifications"
            )
            kwargs["stored_settings"]["_ewoks_default_inputs"] = static_input
        return super().__new__(cls, *args, **kwargs)

    @Inputs.data
    def process(self, scanID, *args, **kwargs):
        if scanID is None:
            return
        assert isinstance(scanID, TomwerScanBase)
        ProcessManager().notify_dataset_state(
            dataset=scanID, process=self, state=DatasetState.PENDING
        )
        configuration = self.widget.recons_params.to_dict()
        configuration["process_only_copy"] = False  # we never only do copy from the gui
        configuration["process_only_dkrf"] = not self.widget.isCopyActive()
        configuration["mode_auto"] = self.widget.isOnModeAuto()

        self._processing_stack.add(copy.copy(scanID), configuration=configuration)

    @Inputs.reduced_darks
    def received_reduced_darks(self, reduced_darks: dict):
        # we consider if the darks are provided then the user want to use those and don't copy them manually
        self.setModeAuto(False)
        reduced_darks.pop("reduce_frames_name", None)
        self.setReducedDarks(reduced_darks)

    @Inputs.reduced_flats
    def received_reduced_flats(self, reduced_flats: dict):
        # we consider if the darks are provided then the user want to use those and don't copy them manually
        self.setModeAuto(False)
        reduced_flats.pop("reduce_frames_name", None)
        self.setReducedFlats(reduced_flats)

    @docstring(SuperviseOW)
    def reprocess(self, dataset):
        self.process(dataset)

    def signalReady(self, scanID):
        assert isinstance(scanID, TomwerScanBase)
        self.Outputs.data.send(scanID)
        self.sigScanReady.emit(scanID)

    def _updateSettingsVals(self):
        self._ewoks_default_inputs = {
            "data": None,
            "dark_ref_params": self.widget.recons_params.to_dict(),
        }
        self._ewoks_default_inputs["dark_ref_params"][
            "auto"
        ] = self.widget.isOnModeAuto()
        self._ewoks_default_inputs["dark_ref_params"][
            "activate"
        ] = self.widget.isCopyActive()

    @property
    def recons_params(self):
        return self.widget.recons_params

    def close(self):
        self.widget.close()
        super(DarkRefAndCopyOW, self).close()

    def _endProcessing(self, scan):
        WidgetLongProcessing._endProcessing(self, scan)
        self.Outputs.data.send(scan)
        self.sigScanReady.emit(scan)
        if scan.reduced_darks not in (None, {}):
            # we want to send those in relative position to have something generic. This is a convention for now
            reduced_darks = scan.reduced_darks
            reduced_darks.pop("reduce_frames_name", None)
            self.Outputs.reduced_darks.send(
                tomoscan.esrf.scan.utils.from_absolute_reduced_frames_to_relative(
                    reduced_frames=reduced_darks, scan=scan
                )
            )
        if scan.reduced_flats not in (None, {}):
            # we want to send those in relative position to have something generic. This is a convention for now
            reduced_flats = scan.reduced_flats
            reduced_flats.pop("reduce_frames_name", None)
            self.Outputs.reduced_flats.send(
                tomoscan.esrf.scan.utils.from_absolute_reduced_frames_to_relative(
                    reduced_frames=reduced_flats, scan=scan
                )
            )
        _logger.info(f"{scan} ended")

    def setReducedDarks(self, darks: dict):
        self.widget._refCopyWidget.save_darks_to_be_copied(darks)

    def setReducedFlats(self, flats: dict):
        self.widget._refCopyWidget.save_flats_to_be_copied(flats)



from __future__ import annotations

import sys
from silx.gui import qt


class RangeWidget(qt.QWidget):
    """widget to select a range of float"""

    sigChanged = qt.Signal()

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setLayout(qt.QHBoxLayout())

        # start
        self.layout().addWidget(qt.QLabel("from"))
        self._startQLE = qt.QDoubleSpinBox(self)
        self.layout().addWidget(self._startQLE)
        self._startQLE.setRange(sys.float_info.min, sys.float_info.max)

        # stop
        self.layout().addWidget(qt.QLabel("to"))
        self._stopQLE = qt.QDoubleSpinBox(self)
        self.layout().addWidget(self._stopQLE)
        self._stopQLE.setRange(sys.float_info.min, sys.float_info.max)

        # connect signal / slot
        self._startQLE.editingFinished.connect(self.sigChanged)
        self._stopQLE.editingFinished.connect(self.sigChanged)

    def getRange(self) -> tuple[float]:
        return (
            self._startQLE.value(),
            self._stopQLE.value(),
        )

    def setRange(self, start: float, stop: float) -> None:
        self._startQLE.setValue(float(start))
        self._stopQLE.setValue(float(stop))

    def setSuffix(self, suffix: str):
        self._startQLE.setSuffix(suffix)
        self._stopQLE.setSuffix(suffix)



# coding: utf-8
from __future__ import annotations

from orangewidget.settings import set_widget_settings_dir_components
from orangewidget.workflow import config
from silx.gui import qt
import logging

import tomwer.version
from tomwer.gui import icons

from ewoksorange.pkg_meta import iter_entry_points
from . import environ
from .splash import getIcon, splash_screen
from .widgetsscheme import WidgetsScheme

_logger = logging.getLogger(__name__)


def version():
    return tomwer.version.version


class TomwerConfig(config.Config):
    """
    Configuration defined for tomwer
    """

    OrganizationDomain = "esrf"
    ApplicationName = "tomwer"
    ApplicationVersion = version()

    def init(self):
        super().init()
        qt.QApplication.setApplicationName(self.ApplicationName)
        qt.QApplication.setOrganizationDomain(self.OrganizationDomain)
        qt.QApplication.setApplicationVersion(self.ApplicationVersion)
        qt.QApplication.setApplicationDisplayName(self.ApplicationName)
        widget_settings_dir_cfg = environ.get_path("widget_settings_dir", "")
        if widget_settings_dir_cfg:
            # widget_settings_dir is configured via config file
            set_widget_settings_dir_components(
                widget_settings_dir_cfg, self.ApplicationVersion
            )

        canvas_settings_dir_cfg = environ.get_path("canvas_settings_dir", "")
        if canvas_settings_dir_cfg:
            # canvas_settings_dir is configured via config file
            qt.QSettings.setPath(
                qt.QSettings.IniFormat, qt.QSettings.UserScope, canvas_settings_dir_cfg
            )

    @staticmethod
    def splash_screen():
        """"""
        return splash_screen()

    @staticmethod
    def core_packages():
        return super().core_packages() + ["tomwer-add-on"]

    @staticmethod
    def application_icon():
        return getIcon()

    @staticmethod
    def workflow_constructor(*args, **kwargs):
        return WidgetsScheme(*args, **kwargs)

    @staticmethod
    def widgets_entry_points():
        """
        Return an `EntryPoint` iterator for all 'orange.widget' entry
        points.
        """
        # Ensure the 'this' distribution's ep is the first. iter_entry_points
        # yields them in unspecified order.
        WIDGETS_ENTRY = "orange.widgets"

        def accept_extension(entry):
            # safer to ignore some extension like ewoks demo
            return entry.name.lower() not in ("ewoks demo",)

        try:
            all_eps = filter(accept_extension, iter_entry_points(group=WIDGETS_ENTRY))
        except Exception as e:
            _logger.error(f"fail to find add-ons. Error is {e}")
        return iter(all_eps)

    @staticmethod
    def addon_entry_points():
        return TomwerConfig.widgets_entry_points()

    APPLICATION_URLS = {
        #: Submit a bug report action in the Help menu
        "Bug Report": "https://gitlab.esrf.fr/tomotools/tomwer/-/issues",
        #: A url quick tour/getting started url
        "Quick Start": "https://www.youtube.com/playlist?list=PLddRXwP6Z6F9KOu1V5o6H24KPH5Ikuk2f",
        #: The 'full' documentation, should be something like current /docs/
        #: but specific for 'Visual Programing' only
        "Documentation": "https://tomotools.gitlab-pages.esrf.fr/tomwer/index.html",
        #: YouTube tutorials
        "Screencasts": "https://www.youtube.com/@tomotools",
    }


class TomwerSplashScreen(qt.QSplashScreen):
    def __init__(
        self,
        parent=None,
        pixmap=None,
        textRect=None,
        textFormat=qt.Qt.PlainText,
        **kwargs,
    ):
        super(TomwerSplashScreen, self).__init__(pixmap=icons.getQPixmap("tomwer"))

    def showMessage(self, message, alignment=qt.Qt.AlignLeft, color=qt.Qt.black):
        version = f"tomwer version {tomwer.version.version}"
        super().showMessage(version, qt.Qt.AlignLeft | qt.Qt.AlignBottom, qt.Qt.white)



# coding: utf-8
from __future__ import annotations

import os

import numpy
import pytest

from tomwer.core.process.reconstruction.nabu.castvolume import CastVolumeTask
from tomwer.core.process.reconstruction.output import NabuOutputFileFormat
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.volume.edfvolume import EDFVolume
from tomwer.core.volume.hdf5volume import HDF5Volume
from tomwer.core.volume.tiffvolume import TIFFVolume


def test_cast_volume_32bitstiff_to_16bits_tiff(tmp_path):
    """test 32 bits tiffs to 16 bits works"""
    dir_test = tmp_path / "test_dir"
    dir_test.mkdir()
    save_dir = tmp_path / "output_dir"
    save_dir.mkdir()

    vol_data = numpy.random.random(20 * 100 * 100) * 3600.4
    vol_data = vol_data.reshape(20, 100, 100)
    vol_data = vol_data.astype(numpy.float32)
    volume = TIFFVolume(
        folder=dir_test,
        volume_basename="test_dir_vol_",
        data=vol_data,
    )
    volume.save()

    # test with output datadir as None
    with pytest.raises(ValueError):
        CastVolumeTask(
            inputs={
                "volume": volume,
                "configuration": {
                    "output_file_path": None,
                    "output_file_format": NabuOutputFileFormat.EDF,
                    "output_data_type": numpy.uint16,
                },
            },
        ).run()

    # test providing save dir and tiff
    assert len(os.listdir(save_dir)) == 0
    task = CastVolumeTask(
        inputs={
            "volume": volume,
            "configuration": {
                "output_dir": str(save_dir),
                "output_file_format": NabuOutputFileFormat.EDF,
                "output_data_type": numpy.uint16,
            },
        }
    )
    task.run()

    assert len(os.listdir(save_dir)) == 20, "no files have been generated"
    assert task.outputs.volume.load_data().shape == (20, 100, 100)
    assert task.outputs.volume.load_data().dtype == numpy.uint16

    # test providing save dir, a scan and hdf5
    scan = NXtomoScan(scan=os.sep.join([str(tmp_path), "scan"]), entry="entry0000")
    task = CastVolumeTask(
        inputs={
            "volume": volume,
            "configuration": {
                "output_dir": str(save_dir),
                "output_file_format": NabuOutputFileFormat.HDF5,
                "output_data_type": numpy.float32,
            },
            "scan": scan,
        },
    )
    task.run()
    assert task.outputs.volume.load_data().shape == (20, 100, 100)
    assert task.outputs.volume.load_data().dtype == numpy.float32

    # test providing a volume as output volume
    output_volume = HDF5Volume(
        file_path=os.path.join(str(tmp_path), "my_volume.hdf5"), data_path="entry0002"
    )
    task = CastVolumeTask(
        inputs={
            "volume": volume,
            "configuration": {
                "output_dir": str(save_dir),
                "output_file_format": NabuOutputFileFormat.HDF5,
                "output_data_type": numpy.uint8,
            },
            "output_volume": output_volume,
        },
    )
    task.run()
    assert task.outputs.volume.load_data().shape == (20, 100, 100)
    assert task.outputs.volume.load_data().dtype == numpy.uint8

    # test providing a volume us as output volume
    output_volume = EDFVolume(folder=os.path.join(str(tmp_path), "my_edf_vol"))
    task = CastVolumeTask(
        inputs={
            "volume": volume,
            "configuration": {
                "output_dir": str(save_dir),
                "output_file_format": NabuOutputFileFormat.HDF5,
                "output_data_type": numpy.int16,
            },
            "output_volume": output_volume.get_identifier().to_str(),
        },
    )
    task.run()
    assert task.outputs.volume.load_data().shape == (20, 100, 100)
    assert task.outputs.volume.load_data().dtype == numpy.int16



# coding: utf-8
from __future__ import annotations


import sys

from silx.gui import qt


class ObjectInspector(qt.QWidget):
    def __init__(self, parent):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QVBoxLayout())

        self.__model = qt.QStandardItemModel()

        self._treeInspectorView = qt.QTreeView(self)
        self._treeInspectorView.setModel(self.__model)
        self.layout().addWidget(self._treeInspectorView)

    def setObject(self, object):
        self.__model.clear()
        root_item = qt.QStandardItem(f"{str(object)}")
        size_item = qt.QStandardItem(f"{sys.getsizeof(object) * 10e-6} (M bytes)")
        self.__model.invisibleRootItem().appendRow((root_item, size_item))

        def add_sub_elmts(parent_item, python_obj):
            if hasattr(python_obj, "__dict__"):
                for i_elmt, elmt in enumerate(python_obj.__dict__):
                    if hasattr(object, elmt):
                        children_obj = getattr(object, elmt)
                        # TODO: check if is a property or is a function
                        item = qt.QStandardItem(f"{elmt}")
                        # TODO: add size
                        size_item = qt.QStandardItem(
                            f"{sys.getsizeof(children_obj) * 10e-6} (M bytes)"
                        )
                        parent_item.appendRow((item, size_item))

                        # TODO: add children
                        add_sub_elmts(parent_item=item, python_obj=children_obj)

        add_sub_elmts(parent_item=root_item, python_obj=object)



from __future__ import annotations

import logging

from silx.gui import qt

from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.process.reconstruction.axis import mode as axis_mode
from tomwer.gui.utils.buttons import PadlockButton
from tomwer.gui.utils.qt_utils import block_signals
from tomwer.synctools.axis import QAxisRP
from tomwer.gui.utils.scrollarea import QComboBoxIgnoreWheel
from tomwer.gui.reconstruction.axis.EstimatedCORWidget import EstimatedCORWidget

_logger = logging.getLogger(__name__)


class CalculationWidget(qt.QWidget):
    """
    Main widget to select the algorithm to use for COR calculation
    Used as a tab of the AxisSettingsTabWidget
    """

    sigModeChanged = qt.Signal(str)
    """signal emitted when the algorithm for computing COR changed"""

    sigLockModeChanged = qt.Signal(bool)
    """signal emitted when the mode has been lock or unlock"""
    sigUpdateXRotAxisPixelPosOnNewScan = qt.Signal()
    sigYAxisInvertedChanged = qt.Signal(bool)

    def __init__(self, parent, axis_params):
        assert isinstance(axis_params, QAxisRP)
        qt.QWidget.__init__(self, parent)
        self._axis_params = None
        self.setLayout(qt.QVBoxLayout())

        # algorithm
        self._modeWidget = qt.QWidget(parent=self)
        self._modeWidget.setLayout(qt.QHBoxLayout())
        self.layout().addWidget(self._modeWidget)

        self.__rotAxisSelLabel = qt.QLabel("algorithm to compute cor")
        self._modeWidget.layout().addWidget(self.__rotAxisSelLabel)
        self._qcbPosition = QComboBoxIgnoreWheel(self)

        algorithm_groups = (
            # radio group
            (
                axis_mode.AxisMode.centered,
                axis_mode.AxisMode.global_,
                axis_mode.AxisMode.growing_window_radios,
                axis_mode.AxisMode.sliding_window_radios,
                axis_mode.AxisMode.octave_accurate_radios,
            ),
            # sino group
            (
                axis_mode.AxisMode.growing_window_sinogram,
                axis_mode.AxisMode.sino_coarse_to_fine,
                axis_mode.AxisMode.sliding_window_sinogram,
                axis_mode.AxisMode.fourier_angles,
            ),
            # composite coarse to fine
            (
                axis_mode.AxisMode.composite_coarse_to_fine,
                axis_mode.AxisMode.near,
            ),
            # read
            (axis_mode.AxisMode.read,),
            # manual
            (axis_mode.AxisMode.manual,),
        )
        current_pos = 0
        for i_grp, algorithm_group in enumerate(algorithm_groups):
            if i_grp != 0:
                self._qcbPosition.insertSeparator(current_pos)
                current_pos += 1
            for cor_algorithm in algorithm_group:
                self._qcbPosition.addItem(cor_algorithm.value)
                idx = self._qcbPosition.findText(cor_algorithm.value)
                self._qcbPosition.setItemData(
                    idx,
                    axis_mode.AXIS_MODE_METADATAS[cor_algorithm].tooltip,
                    qt.Qt.ToolTipRole,
                )
                current_pos += 1

        self._modeWidget.layout().addWidget(self._qcbPosition)

        # method lock button
        self._lockMethodPB = PadlockButton(parent=self._modeWidget)
        self._lockMethodPB.setToolTip(
            "Lock the method to compute the cor. \n"
            "This will automatically call the "
            "defined algorithm each time a scan is received."
        )
        self._modeWidget.layout().addWidget(self._lockMethodPB)

        # estimated cor
        self._estimatedCorWidget = EstimatedCORWidget(self, axis_params=axis_params)
        self.layout().addWidget(self._estimatedCorWidget)

        # connect signal / slot
        self._qcbPosition.currentIndexChanged.connect(self._modeChanged)
        self._lockMethodPB.sigLockChanged.connect(self.lockMode)
        self._estimatedCorWidget.sigUpdateXRotAxisPixelPosOnNewScan.connect(
            self.sigUpdateXRotAxisPixelPosOnNewScan
        )
        self._estimatedCorWidget.sigYAxisInvertedChanged.connect(
            self.sigYAxisInvertedChanged
        )

        # set up interface
        self._estimatedCorWidget._updateVisibleSides(mode=self.getMode())
        self.setAxisParams(axis_params)

    def getMethodLockPB(self) -> qt.QPushButton:
        return self._lockMethodPB

    def setEstimatedCorValue(self, value):
        self._estimatedCorWidget.setEstimatedCor(value=value)
        # note: force to update the side values.
        self._estimatedCorWidget._updateVisibleSides(mode=self.getMode())

    def getEstimatedCor(self):
        return self._estimatedCorWidget.getEstimatedCor()

    def updateXRotationAxisPixelPositionOnNewScan(self) -> bool:
        return self._estimatedCorWidget.updateXRotationAxisPixelPositionOnNewScan()

    def setUpdateXRotationAxisPixelPositionOnNewScan(self, update: bool):
        self._estimatedCorWidget.setUpdateXRotationAxisPixelPositionOnNewScan(
            update=update
        )

    def _modeChanged(self, *args, **kwargs):
        mode = self.getMode()
        with block_signals(self._qcbPosition):
            with block_signals(self._axis_params):
                self._estimatedCorWidget._updateVisibleSides(mode)
                self._axis_params.mode = mode.value
            self._axis_params.changed()
            self.sigModeChanged.emit(mode.value)

    def isModeLock(self):
        return self._lockMethodPB.isLocked()

    def setModeLock(self, mode=None):
        """set a specific mode and lock it.

        :param mode: mode to lock. If None then keep the current mode
        """
        if mode is not None:
            mode = axis_mode.AxisMode.from_value(mode)
        if mode is None and axis_mode.AXIS_MODE_METADATAS[self.getMode()].is_lockable():
            raise ValueError(
                "Unable to lock the current mode is not an automatic algorithm"
            )
        elif (
            mode != self.getMode() and axis_mode.AXIS_MODE_METADATAS[mode].is_lockable()
        ):
            raise ValueError("Unable to lock %s this is not a lockable mode")

        if mode is not None:
            self.setMode(mode)
        if not self._lockMethodPB.isLocked():
            with block_signals(self._lockMethodPB):
                self._lockMethodPB.setLock(True)
        self.lockMode(True)

    def lockMode(self, lock):
        with block_signals(self._lockMethodPB):
            self._lockMethodPB.setLock(lock)
            self._qcbPosition.setEnabled(not lock)

        self.sigLockModeChanged.emit(lock)

    def getMode(self):
        """Return algorithm to use for axis calculation"""
        return axis_mode.AxisMode.from_value(self._qcbPosition.currentText())

    def setMode(self, mode: axis_mode.AxisMode):
        if mode == self.getMode():
            # when set mode the estimated cor might changed.
            # that we want to avoid.
            return
        with block_signals(self._qcbPosition):
            index = self._qcbPosition.findText(mode.value)
            if index >= 0:
                self._qcbPosition.setCurrentIndex(index)
            else:
                raise ValueError("Unable to find mode", mode)
            self._lockMethodPB.setVisible(mode not in (axis_mode.AxisMode.manual,))
            mode_metadata = axis_mode.AXIS_MODE_METADATAS[mode]
            estimated_cor_widget_visible = (
                mode_metadata.allows_estimated_cor_as_numerical_value
                or len(mode_metadata.valid_sides) > 0
            )
            self._estimatedCorWidget.setVisible(estimated_cor_widget_visible)
            self._estimatedCorWidget._updateVisibleSides(mode=mode)

    def setAxisParams(self, axis):
        with block_signals(self):
            if self._axis_params is not None:
                self._axis_params.sigChanged.disconnect(self._axis_params_changed)
            self._axis_params = axis
            if self._axis_params.mode in (axis_mode.AxisMode.manual,):
                # those mode cannot be handled by the auto calculation dialog
                self._axis_params.mode = axis_mode.AxisMode.growing_window_radios
            self._axis_params.sigChanged.connect(self._axis_params_changed)
            self._axis_params_changed()

    def _axis_params_changed(self, *args, **kwargs):
        self.setMode(self._axis_params.mode)

    def setScan(self, scan: TomwerScanBase | None):
        self._estimatedCorWidget.setPixelSize(pixel_size_m=scan.x_pixel_size)
        self._estimatedCorWidget.setImageWidth(image_width=scan.dim_1)



# coding: utf-8

try:
    from silx.gui import qt
except ImportError:
    raise ImportError("Can't found silx modules")
import tomwer.version
from tomwer.gui import icons


def getMainSplashScreen():
    pixmap = icons.getQPixmap("tomwer")
    splash = qt.QSplashScreen(pixmap)
    splash.show()
    splash.raise_()
    _version = tomwer.version.version
    text = "version " + str(_version)
    splash.showMessage(text, qt.Qt.AlignLeft | qt.Qt.AlignBottom, qt.Qt.white)
    return splash
